import os
from pathlib import Path
from typing import List, Tuple, Union
from uuid import UUID, uuid4

from ....client.io import _pil_open_image_path
from ....core import OBJECT_CLASS_ATTRIBUTE_UUID
from ...base_models import AttributeRecord, ImageRecord
from ...interfaces import IReader

PathLike = Union[str, Path]


def is_uuid(s):
    try:
        _ = UUID(s)
        return True
    except ValueError as e:
        return False


def get_class_dir_uuids(root_dir):
    uuids = [
        item
        for item in os.listdir(root_dir)
        if (os.path.isdir(os.path.join(root_dir, item)) and is_uuid(item))
    ]
    return uuids


class TorchImageFolderReader(IReader):
    format_name = "torch-image-folder"

    def __init__(
        self,
        root_dir: PathLike,
        attribute_id: str = str(OBJECT_CLASS_ATTRIBUTE_UUID),
        attribute_name: str = OBJECT_CLASS_ATTRIBUTE_UUID.label,
    ):
        """Read a ImageFolder classification dataset from disk into
        highlighter Dataset format

        root_dir: Root dir of classification dataset. This dir should
        contain dirs named for theHighlighterEntityAttributeEnumID used for
        the classification task.

        attribute_id: Default OBJECT_CLASS_ATTRIBUTE_UUID
        attribute_name: Default 'object_class'

        """
        self.root_dir = str(root_dir)
        self.attribute_id = attribute_id
        self.attribute_name = attribute_name

    def read(self) -> Tuple[List[AttributeRecord], List[ImageRecord]]:
        class_uuids = get_class_dir_uuids(self.root_dir)

        attribute_records = []
        data_file_records = []
        for value in class_uuids:
            image_filenames = os.listdir(os.path.join(self.root_dir, value))
            for filename in image_filenames:
                data_file_id = Path(filename).stem
                w, h = _pil_open_image_path(os.path.join(self.root_dir, self.attribute_id, filename)).size

                data_file_records.append(
                    ImageRecord(
                        data_file_id=data_file_id, width=w, height=h, filename=filename, extra_fields={}
                    )
                )

                attribute_records.append(
                    AttributeRecord(
                        data_file_id=data_file_id,
                        entity_id=str(uuid4()),
                        attribute_id=self.attribute_id,
                        attribute_name=self.attribute_name,
                        value=value,
                    )
                )

        return data_file_records, attribute_records



# Usage
# ~~~~~
# Can't use utilities.logger.LoggingHandlerMQTT to diagnose this MQTT module !
#
# When using AIKO_LOG_LEVEL=DEBUG for the rest of framework or applications,
# typically don't want to see low-level MQTT message debug log records.
# Therefore, AIKO_LOG_LEVEL=DEBUG doesn't affect this MQTT module, instead use
#
#     AIKO_LOG_LEVEL_MQTT=DEBUG application parameters ...
#
# To Do
# ~~~~~
# - Allow default MQTT_HOST and MQTT_PORT to be overridden by CLI parameters
#
# - BUG: If on MQTT thread, then if waiting for a condition to change that
#   depends upon an incoming message, then wait will timeout, because that
#   incoming message can't be processed.
#   For example, if wait_disconnected() called whilst on MQTT thread,
#   then the MQTT broker message for "disconnect" --> on_disconnect()
#   won't happen.
#   For example, when Registrar processed "(primary absent)" message and
#   attempts set_last_will_and_testament(), which causes a wait_disconnected()
#   whilst on the MQTT thread.
# - SOLUTION: By default queue all incoming MQTT messages and process on the
#   main event.loop() !
#
# - Track which topic subscriptions have succeeded ...
#     mqtt_client.subscribe() return value is saved (subscribe topic list index)
#     to be compared to the on_subscribe(..., mid, ...) parameter value
#
# - Refactor wait_disconnected(), wait_connected() and wait_published(self)
#     into a single function
#   - Parameter is a variable condition to test,
#       e.g not self.connected, self.connected and self.published
#
# - Implement message to change logging level !
# - Implement reconnect on disconnection from MQTT server
#   - Maybe raise an exception if reconnection fails, rather than sys.exit(-1)
# - Implement discovery and manual setting of AIKO_MQTT_HOST, etc
#   - Rediscover on disconnection
# - Implement destructor ?

import os
import paho.mqtt.client as mqtt
import socket
import time
from typing import Any, List

from aiko_services.main.message import *
from aiko_services.main.utilities import *

__all__ = ["MQTT"]

# MQTT_HOST = "mqtt.fluux.io"
# MQTT_HOST = "test.mosquitto.org"

_AIKO_LOG_LEVEL = os.environ.get("AIKO_LOG_LEVEL_MQTT", "INFO")
_LOGGER = get_logger(__name__, log_level=_AIKO_LOG_LEVEL, logging_handler=None)
_MAXIMUM_WAIT_TIME = 2000  # milliseconds

def _on_message(mqtt_client: Any, userdata: Any, message: Any) -> None:
    if _LOGGER.isEnabledFor(DEBUG):  # Don't expand debug message
        _LOGGER.debug(f"message: {message.topic}: {message.payload}")

class MQTT(Message):
    def __init__(
        self: Any,
        message_handler: Any = _on_message,
        topics_subscribe: Any = None,
        topic_lwt: str = None,
        payload_lwt:str = None,
        retain_lwt: bool = False,
        mqtt_state_handler = None
        ) -> None:

        self.message_handler = message_handler
        self.mqtt_state_handler = mqtt_state_handler
        self.connected = MessageState.DISCONNECTED
        if self.mqtt_state_handler:
            self.mqtt_state_handler(self.connected)
        self.published = True
        self.topics_subscribe = []
        self.wildcard_topic = False
        self.wildcard_subscribed = False
        self.subscribe(topics_subscribe)

        mqtt_configuration = get_mqtt_configuration()
        server_up = mqtt_configuration[0]
        self.mqtt_host = mqtt_configuration[1]
        self.mqtt_port = mqtt_configuration[2]
        self.mqtt_transport = mqtt_configuration[3]
        self.mqtt_username = mqtt_configuration[4]
        self.mqtt_password = mqtt_configuration[5]
        self.mqtt_tls_enabled = mqtt_configuration[6]
        tls_state = "TLS enabled" if self.mqtt_tls_enabled else "TLS disabled"
        self.mqtt_info = f"{self.mqtt_host}:{self.mqtt_port}:{tls_state}"

        if server_up:
            self._connect(topic_lwt, payload_lwt, retain_lwt)
        else:
            diagnostic = f"Couldn't connect to MQTT server {self.mqtt_info}"
            raise SystemError(diagnostic)

    def _connect(
        self: Any,
        topic_lwt: Any,
        payload_lwt: str,
        retain_lwt: Any
        ) -> None:

        _LOGGER.debug(f"connecting to {self.mqtt_info}")
        self.mqtt_client = mqtt.Client(transport=self.mqtt_transport)
        self.mqtt_client.on_connect = self._on_connect
        self.mqtt_client.on_disconnect = self._on_disconnect
        self.mqtt_client.on_message = self.message_handler
        self.mqtt_client.on_publish = self._on_publish

        if topic_lwt:
            self.mqtt_client.will_set(
                topic_lwt, payload=payload_lwt, retain=retain_lwt)
        if self.mqtt_tls_enabled:
            self.mqtt_client.tls_set()
        if self.mqtt_username:
            self.mqtt_client.username_pw_set(
                self.mqtt_username, self.mqtt_password)

        diagnostic = f"Couldn't connect to MQTT server {self.mqtt_info}"
        try:
            self.mqtt_client.connect(
                host=self.mqtt_host, port=self.mqtt_port, keepalive=60)
            self.mqtt_client.loop_start()  # Handles MQTT reconnections
        except socket.gaierror:
            raise SystemError(diagnostic)
        except ConnectionRefusedError:
            raise SystemError(diagnostic)
    #   except ConnectionResetError:  # TODO: Is this required as well ?
    #       raise SystemError(diagnostic)

    def _disconnect(self: Any) -> None:
        _LOGGER.debug(f"disconnect from {self.mqtt_info}")
        self.wait_published()
        self.mqtt_client.loop_stop()
        self.mqtt_client.disconnect()    # Note: Does not cause LWT to be sent
#       self.mqtt_client.loop_forever()  # Handles MQTT reconnections
        self.mqtt_client = None

    # pylint: disable=unused-argument
    def _on_connect(
        self: Any,
        mqtt_client: Any,
        user_data: Any,
        dict_flags: Any,
        result_code: Any
        ) -> None:

        _LOGGER.debug(f"connected to {self.mqtt_info}")
        self.connected = MessageState.CONNECTED
        if self.mqtt_state_handler:
            self.mqtt_state_handler(self.connected)
        self._subscribe_if_connected(self.topics_subscribe)

    def _on_disconnect(
        self: Any,
        mqtt_client: Any,
        user_data: Any,
        return_code: Any
        ) -> None:

        _LOGGER.debug(f"on_disconnect")
        self.connected = MessageState.DISCONNECTED
        if self.mqtt_state_handler:
            self.mqtt_state_handler(self.connected)
        if return_code != 0:
            _LOGGER.info(f"on_disconnect: will reconnect: {return_code}")

    # pylint: disable=unused-argument
    def _on_publish(
        self: Any,
        mqtt_client: Any,
        userdata: Any,
        result: Any
        ) -> None:

        self.published = True

    def publish(
        self: Any,
        topic: str,
        payload: Any,
        retain: bool = False,
        wait: bool = False
        ) -> None:

        self.published = False
        self.wait_connected()
        self.mqtt_client.publish(topic, payload, retain=retain)
        if wait:
            self.wait_published()

    def set_last_will_and_testament(
        self: Any,
        topic_lwt: str = None,
        payload_lwt: str = "(absent)",
        retain_lwt: bool = False
        ) -> None:

        self._disconnect()
        self.wait_disconnected()
        self._connect(topic_lwt, payload_lwt, retain_lwt)

    def subscribe(self: Any, topics: Any) -> None:
        if topics:
            if isinstance(topics, str):
                topics = [topics]
            if isinstance(topics, dict):
                topics = topics.keys()
            for topic in topics:
                if topic == "#":
                    self.wildcard_topic = True
                    self.unsubscribe(self.topics_subscribe, remove=False)
                else:
                    self.topics_subscribe.append(topic)

            self._subscribe_if_connected(topics)

    def _subscribe_if_connected(self, topics: Any):
        if self.connected == MessageState.CONNECTED:
            if self.wildcard_topic:
                if not self.wildcard_subscribed:
                    self.mqtt_client.subscribe("#")
                    self.wildcard_subscribed = True
                    _LOGGER.debug(f"subscribed to {self.mqtt_info}: #")
            elif topics:
                for topic in topics:
                    self.mqtt_client.subscribe(topic)
                    _LOGGER.debug(f"subscribed to {self.mqtt_info}: {topic}")

    def unsubscribe(self: Any, topics: Any, remove=True) -> None:
        if topics:
            if isinstance(topics, str):
                topics = [topics]
            if isinstance(topics, dict):
                topics = topics.keys()
            for topic in topics:
                if topic == "#":
                    if self.wildcard_topic:
                        self.wildcard_topic = False
                        if self.wildcard_subscribed:
                            self.mqtt_client.unsubscribe("#")
                            self.wildcard_subscribed = False
                            _LOGGER.debug(f"unsubscribed from {self.mqtt_info}: #")
                        self._subscribe_if_connected(self.topics_subscribe)
                elif topic in self.topics_subscribe:
                    if remove:
                        self.topics_subscribe.remove(topic)
                    if self.connected == MessageState.CONNECTED:
                        self.mqtt_client.unsubscribe(topic)
                        _LOGGER.debug(f"unsubscribed from {self.mqtt_info}: {topic}")

# TODO: Only wait a limited time and either carry on without failing ...
#       or fail and choose to reconnect to MQTT

    def wait_disconnected(self: Any) -> None:
#       _LOGGER.debug("wait disconnected")
        for counter in range(_MAXIMUM_WAIT_TIME + 1):
            if self.connected == MessageState.DISCONNECTED:
                break
            time.sleep(0.001)
        if counter >= _MAXIMUM_WAIT_TIME:
            _LOGGER.error(f"wait disconnected timeout: {counter}")
#       else:
#           _LOGGER.debug(f"wait disconnected: {counter}")

# TODO: Only wait a limited time and either carry on without failing ...
#       or fail and choose to reconnect to MQTT

    def wait_connected(self: Any) -> None:
#       _LOGGER.debug("wait connected")
        for counter in range(_MAXIMUM_WAIT_TIME + 1):
            if self.connected == MessageState.CONNECTED:
                break
            time.sleep(0.001)
        if counter >= _MAXIMUM_WAIT_TIME:
            _LOGGER.error(f"wait connected timeout: {counter}")
#       else:
#           _LOGGER.debug(f"wait connected: {counter}")

# TODO: Only wait a limited time and either carry on without failing ...
#       or fail and choose to reconnect to MQTT

    def wait_published(self: Any) -> None:
#       _LOGGER.debug("wait published")
        for counter in range(_MAXIMUM_WAIT_TIME + 1):
            if self.published: break
            time.sleep(0.001)
        if counter >= _MAXIMUM_WAIT_TIME:
            _LOGGER.error(f"wait published timeout: {counter}")
#       else:
#           _LOGGER.debug(f"wait published: {counter}")



from typing import TypeVar
from uuid import UUID

from sqlmodel import Field, select

T = TypeVar("T", bound="AccountMixin")


class AccountMixin:
    account_uuid: UUID = Field(index=True)

    @classmethod
    def base_query(cls, account_uuid: UUID):
        return select(cls).where(cls.account_uuid == account_uuid)



# Example
# ~~~~~~~
# from aiko_services.main.proxy_play import Proxy, ProxyExample
# class C():
#     def f(self, argument):
#         print(f"C.f({argument})")
# proxy_C = proxy.Proxy(C())
# proxy_C.f(1)
# --> C.f(1)
#
# intercepted_C = proxy.ProxyExample(C())
# intercepted_C.f(1, "2", 3, a=4, b=5)
# --> Intercepted: name: f, args: (1, "2", 3), kwargs: {"a": 4, "b": 5}
#
# Issues
# ~~~~~~
# Proxying user-types should work okay
# However, proxying builtin objects (ints, floats, lists, etc),
# has some limitation and inconsistencies, imposed by the interpreter
#
# Currently, can't use Proxy object as a dictionary key (rare case)
#
#   class A(object):
#       pass
#   p = proxy.Proxy(A)
#   {p: "asdasd"}
#   --> TypeError: unhashable type: "Proxy(type)"
#
# Resources
# ~~~~~~~~~
# Covered by https://creativecommons.org/licenses/by-sa/3.0
# - https://code.activestate.com/help/terms
# - https://code.activestate.com/recipes/496741-object-proxying
# - https://code.activestate.com/recipes/578014-lazy-load-object-proxying

from inspect import isfunction

__all__ = ["Proxy"]


class Proxy(object):
    intercept = None  # TODO: More flexible, if this isn't a class variable ?

    __slots__ = ["_obj", "__weakref__"]

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)

    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __hash__(self):
        return hash(object.__getattribute__(self, "_obj"))

    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_obj"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)

    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))
    #   return unicode(self).encode("utf-8")

#   def __unicode__(self):
#       return unicode(object.__getattribute__(self, "_obj"))

    _methods_to_wrap = [
        "__abs__", "__add__", "__and__", "__call__", "__cmp__", "__coerce__",
        "__contains__", "__delitem__", "__delslice__", "__div__", "__divmod__",
        "__eq__", "__float__", "__floordiv__", "__ge__", "__getitem__",
        "__getslice__", "__gt__", "__hex__", "__iadd__", "__iand__",
        "__idiv__", "__idivmod__", "__ifloordiv__", "__ilshift__", "__imod__",
        "__imul__", "__int__", "__invert__", "__ior__", "__ipow__",
        "__irshift__", "__isub__", "__iter__", "__itruediv__", "__ixor__",
        "__le__", "__len__", "__long__", "__lshift__", "__lt__", "__mod__",
        "__mul__", "__ne__", "__neg__", "__oct__", "__or__", "__pos__",
        "__pow__", "__radd__", "__rand__", "__rdiv__", "__rdivmod__",
        "__reduce__", "__reduce_ex__", "__repr__", "__reversed__",
        "__rfloorfiv__", "__rlshift__", "__rmod__", "__rmul__", "__ror__",
        "__rpow__", "__rrshift__", "__rshift__", "__rsub__", "__rtruediv__",
        "__rxor__", "__setitem__", "__setslice__", "__sub__", "__truediv__",
        "__xor__", "next"
    ]

    @classmethod
    def _create_class_proxy(cls, user_class):
        """Creates a proxy for the given class"""

        def make_method(name):
            def method(self, *args, **kwargs):
                user_instance = object.__getattribute__(self, "_obj")
                return getattr(user_instance, name)(*args, **kwargs)
            return method

        namespace = {}
        for name in cls._methods_to_wrap:
            if hasattr(user_class, name):
                namespace[name] = make_method(name)
        return type(f"{cls.__name__}({user_class.__name__})", (cls,), namespace)

    @classmethod
    def _update_instance_intercept(cls, user_class, user_instance):
        """For proxy instance, every user defined function is intercepted"""

        def make_method(name):
            def method(*args, **kwargs):
                return cls.intercept(name, *args, **kwargs)
            return method

        if cls.intercept:
            for name in dir(user_instance):
                if isfunction(getattr(user_class, name)):  # user defined ?
                    user_instance.__setattr__(name, make_method(name))

    def __new__(cls, obj, *args, **kwargs):
        """
        Creates an proxy instance referencing `obj`.
        (obj, *args, **kwargs) are passed to this class' __init__,
        so deriving classes can define an __init__ method of their own.
        Note: _class_proxy_cache is unique per deriving class
        (each deriving class must hold its own cache)
        """
        user_class = obj.__class__
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[user_class]
        except KeyError:
            cache[user_class] = theclass = cls._create_class_proxy(user_class)

        user_instance = object.__new__(theclass)
        theclass.__init__(user_instance, obj, *args, **kwargs)
        cls._update_instance_intercept(user_class, user_instance)
        return user_instance


class ProxyExample(Proxy):
    def intercepted(name, *args, **kwargs):
        print(f"Intercepted: name: {name}, args: {args}, kwargs: {kwargs}")

    intercept = intercepted



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# ./video_example.py -if input_filename      -of output_filename  \
#   -r width height -f framerate [-cv]
#
# ./video_example.py -is input_hostname:port -os output_hostname:port  \
#   -r width height -f framerate -cv  [--RTP]
#
# ./video_example.py -if input_filename      -os output_hostname:port  \
#   -r width height -f framerate -cv
#
# ./video_example.py -is input_hostname:port -of output_filename  \
#   -r width height -f framerate -cv  [--RTP]
#
# ./video_example.py -if ../data/football_test_0.mp4 -os 192.168.1.65:5000  \
#   -r 1280 720 -f 30/1 -cv
#
# CAMERA_IP=192.168.1.89:554
# CAMERA_IP=192.168.1.105:554
# TARGET_IP=192.168.1.155:5000
#
# Default is RTSP protocol, use "--RTP" option to use RTP protocol
# ./video_example.py -is $CAMERA_IP -os $TARGET_IP -r 640 480 -f 25/1 -cv
#
# To Do
# ~~~~~
# - Turn this into a CI test !

import argparse
import sys

from aiko_services.elements.gstreamer import *

# --------------------------------------------------------------------------- #

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("-if", "--input_filename", type=str, default="",
    help="Video input filename")
  parser.add_argument("-is", "--input_stream", type=str, default="",
    help="Video input hostname:port")
  parser.add_argument("-of", "--output_filename", type=str, default="",
    help="Video output filename")
  parser.add_argument("-os", "--output_stream", type=str, default="",
    help="Video output hostname:port")
  parser.add_argument("-r", "--resolution", type=str, default="", nargs=2,
    help="Video resolution width and height")
  parser.add_argument("-f", "--framerate", type=str, default="",
    help="Video resolution width and height")
  parser.add_argument("-cv", "--opencv",
    action='store_true', help="Display on local display and overlay frame id")
  parser.add_argument("--RTP",
    action='store_true', help="Use RTP protocol")
  args = parser.parse_args()

  width, height = args.resolution

  if args.opencv: utilities.enable_opencv()

  if (args.input_filename == "" and args.input_stream == "") or  \
     (args.input_filename != "" and args.input_stream != ""):
    print("usage: Must provide one of either input filename or hostname:port")
    sys.exit(-1)

  if (args.output_filename == "" and args.output_stream == "") or  \
     (args.output_filename != "" and args.output_stream != ""):
    print("usage: Must provide one of either output filename or hostname:port")
    sys.exit(-1)

  try:
    if args.input_filename != "":
      video_reader = VideoFileReader(args.input_filename, width, height)
    else:
      try:
        input_hostname, input_port = args.input_stream.split(":")
      except ValueError as exception:
        print("Error: please specify input as hostname:port")
        sys.exit(-1)
      video_reader = VideoStreamReader(input_hostname, input_port, width, height, rtp=args.RTP)

    if args.output_filename != "":
      video_writer = VideoFileWriter(args.output_filename, width, height, args.framerate)
    else:
      try:
        output_hostname, output_port = args.output_stream.split(":")
      except ValueError as exception:
        print("Error: please specify output as hostname:port")
        sys.exit(-1)
      video_writer = VideoStreamWriter(output_hostname, output_port, width, height, args.framerate)

    utilities.process_video(video_reader, video_writer)

  except ValueError as exception:
    print("Error: " + str(exception))

  except GStreamerError as exception:           # TODO: Needs better diagnostic
    print("GStreamer Error: " + str(exception))

# --------------------------------------------------------------------------- #



# To Do
# ~~~~~
# - Fix ... paths.append((path, None))  # "None" --> should be "file_id" ?

import os
from pathlib import Path

import aiko_services as aiko

__all__ = ["DataSchemeFile"]

# --------------------------------------------------------------------------- #
# parameter: "data_sources" provides the file pathname(s)
# - "(file://pathname_0 file://pathname_1 ...)"
# - "(file://data_in/in_{}.txt)"
# - "(file://data_in/in_{}.jpeg)"
# - "(file://data_in/in_{}.mp4)"
#
# parameter: "data_targets" provides the file pathname(s)
# - "(file://pathname_0 file://pathname_1 ...)"
# - "(file://data_out/out{}.txt)"
# - "(file://data_out/out_{}.jpeg)"
# - "(file://data_out/out_{}.mp4)"

class DataSchemeFile(aiko.DataScheme):
    def create_sources(self,
        stream, data_sources, frame_generator=None, use_create_frame=False):

        if not frame_generator:
            frame_generator = self.frame_generator

        paths = []
        for data_source in data_sources:
            path = aiko.DataScheme.parse_data_url_path(data_source)
            file_glob = "*"
            if aiko.DataScheme.contains_all(path, "{}"):
                file_glob = os.path.basename(path).replace("{}", "*")
                path = os.path.dirname(path)

            path = Path(path)
            if not path.exists():
                diagnostic = f'path "{path}" does not exist'
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

            if path.is_file():
                paths.append((path, None))  # "None" --> should be "file_id" ?
            elif path.is_dir():
                paths_sorted = sorted(path.glob(file_glob))
                file_ids = []
                for path in paths_sorted:
                    file_id = None
                    if file_glob != "*":
                        file_id = self._file_glob_difference(
                            file_glob, path.name)
                    file_ids.append(file_id)
                paths.extend(zip(paths_sorted, file_ids))
            else:
                diagnostic = f'"{path}" must be a file or a directory'
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        pipeline_element = self.pipeline_element
        if use_create_frame and len(paths) == 1:
            pipeline_element.create_frame(stream, {"paths": [path]})
        else:
            stream.variables["source_paths_generator"] = iter(paths)
            rate, _ = pipeline_element.get_parameter("rate", default=None)
            rate = float(rate) if rate else None
            pipeline_element.create_frames(stream, frame_generator, rate=rate)
        return aiko.StreamEvent.OKAY, {}

    def create_targets(self, stream, data_targets):
        path = aiko.DataScheme.parse_data_url_path(data_targets[0])
        target_path_template = aiko.DataScheme.contains_all(path, "{}")

        stream.variables["target_file_id"] = 0
        stream.variables["target_path"] = path
        stream.variables["target_path_template"] = target_path_template
        return aiko.StreamEvent.OKAY, {}

    def _file_glob_difference(self, file_glob, filename):
        difference = None
        tokens = file_glob.split("*")
        token_start = tokens[0]
        token_end = tokens[1] if len(tokens) > 1 else ""
        if filename.startswith(token_start) and filename.endswith(token_end):
            difference = filename[len(token_start):len(filename)-len(token_end)]
        return difference

    def frame_generator(self, stream, frame_id):
        data_batch_size, _ = self.pipeline_element.get_parameter(
            "data_batch_size", default=1)
        data_batch_size = int(data_batch_size)

        paths = []
        try:
            while (data_batch_size > 0):
                data_batch_size -= 1
                path, file_id = next(stream.variables["source_paths_generator"])
                path = Path(path)
                if not path.is_file():
                    diagnostic = f'path "{path}" must be a file'
                    return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
                paths.append(path)
        except StopIteration:
            pass

        if len(paths):
            return aiko.StreamEvent.OKAY, {"paths": paths}
        else:
            return aiko.StreamEvent.STOP, {"diagnostic": "All frames generated"}

aiko.DataScheme.add_data_scheme("file", DataSchemeFile)

# --------------------------------------------------------------------------- #



from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_validator

from ..client import AttributeValue
from ..core import LabeledUUID

__all__ = [
    "AttributeRecord",
    "ImageRecord",
    "S3Files",
]

DEFAULT_SPLIT_NAME = "data"
DEFAULT_DATA_FILES_KEY = "images"  # TODO: Update to 'files'
DEFAULT_ANNOS_KEY = "annotations"  # TODO: Update to 'attributes'
CLOUD_FILES_INFO_KEY = "cloud_files_info"


class ImageRecord(BaseModel):
    data_file_id: Union[int, str]
    width: Optional[int] = None
    height: Optional[int] = None
    filename: str
    split: Optional[str] = DEFAULT_SPLIT_NAME
    extra_fields: Optional[Dict] = None
    assessment_id: Optional[int] = None
    hash_signature: Optional[str] = None


# Extra fields are ignored and will be dropped when instantiating
# a the Object. This is done so things don't break if we load
# some older S3Files where 'type' is a field. If you plan on changing
# this you should consider what to do in this situation
class S3Files(BaseModel, extra="ignore"):
    """Information needed to donwload files/records from s3

    s3://my-bucket/datasets/123/  <-- bucket_name = my-bucket
        records_abc123.json       <-- records_prefix = datasets/123/records_abc123.json
        files_def456.tar.gz       <-- files_prefix = [datasets/123/files_def456.tar.gz,]
    """

    bucket_name: str
    prefix: str
    files: List[str]
    records: List[str] = []
    other: List = []


class AttributeRecord(BaseModel):
    data_file_id: Union[int, str]
    entity_id: UUID = Field(default_factory=uuid4)
    attribute_id: Optional[Union[LabeledUUID, UUID, str]] = None
    attribute_name: str
    value: Any = None
    confidence: Optional[float] = 1.0
    frame_id: Optional[int] = None

    time: Union[datetime, str] = Field(default_factory=datetime.now)
    pipeline_element_name: Optional[str] = None
    training_run_id: Optional[int] = None

    @field_validator("time")
    @classmethod
    def validate_isofrmat(cls, v):
        if isinstance(v, str):
            return datetime.fromisoformat(v)
        else:
            return v

    @classmethod
    def from_attribute_value(
        cls,
        data_file_id: Union[int, str],
        attribute_value: AttributeValue,
        entity_id: Optional[UUID] = None,
        frame_id: Optional[int] = None,
        time: Optional[Union[datetime, str]] = None,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
    ):
        attribute_id = attribute_value.attribute_id
        attribute_name = getattr(attribute_value.attribute_id, "label", "")
        confidence = attribute_value.confidence
        value = attribute_value.value
        return cls(
            data_file_id=data_file_id,
            entity_id=uuid4() if entity_id is None else entity_id,
            attribute_id=attribute_id,
            attribute_name=attribute_name,
            value=value,
            confidence=confidence,
            frame_id=frame_id,
            time=datetime.now() if time is None else time,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
        )

    def dict(self, *args, **kwargs):
        _d = super().model_dump(*args, **kwargs)
        _d["time"] = _d["time"].isoformat()
        return _d

    def to_df_record(self):
        _d = self.dict(exclude_none=True)
        record = {
            "data_file_id": _d.pop("data_file_id"),
            "entity_id": str(_d.pop("entity_id")),
            "attribute_id": str(_d.pop("attribute_id")),
            "attribute_name": _d.pop("attribute_name"),
            "value": _d.pop("value"),
            "confidence": _d.pop("confidence"),
        }
        record["extra_fields"] = _d
        return record



import json
from typing import Dict, Tuple, Union

from highlighter.agent.capabilities import Capability, StreamEvent
from highlighter.client.json_tools import find_valid_json

__all__ = ["ParseJsonFromText"]


class ParseJsonFromText(Capability):
    """
    Parse json from the provided text
    """

    def process_frame(self, stream, text: str) -> Tuple[StreamEvent, Union[Dict, str]]:
        """

        text: str

        """
        try:
            results = find_valid_json(text)

            if len(results) == 0:
                text = "null"
            else:
                text = json.dumps(results[0])

        except json.decoder.JSONDecodeError as e:
            self.logger.error(f"LLM: Error calling find_valid_json on {text}: {e}")
            results = {}

        return StreamEvent.OKAY, {"text": text, "input_text": text}



#!/usr/bin/env python3
#
# Aiko Service: Process Manager
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Create, List(Read), Update and Destroy processes
#
# Usage
# ~~~~~
# ./process_manager.py run  [--name name] [--watchdog] [definition_pathname]
# ./process_manager.py dump [--name name]
# ./process_manager.py exit [--name name] [--grace_time seconds]
#
# ./process_manager.py create  [--name name] [--uid uid] command [arguments ...]
# ./process_manager.py list    [--name name] [uid]
# ./process_manager.py destroy [--name name] [--kill] uid
#
# ./process_manager.py enable | disable | start | status | stop hyperspace_path
#
# Shell and Python scripts
# ~~~~~~~~~~~~~~~~~~~~~~~~
# aiko_process create uid:0 date
# aiko_process create uid:1 sleep 5
# aiko_process create uid:2 aiko_process exit -gt 0  # self terminate !
#
# Simultaneous processes terminating at different times
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# aiko_process create uid:0 /bin/sh -c "echo Start A; sleep 10; echo Stop A"
# aiko_process create uid:0 /bin/sh -c "echo Start B; sleep 15; echo Stop B"
# aiko_process create uid:0 /bin/sh -c "echo Start C; sleep  5; echo Stop C"
#
# Design notes
# ~~~~~~~~~~~~
# Concurrency: Except _run(), all functions execute on the Actor main thread.
# Meaning that those functions are safely executed in order, one-at-a-time.
# The _run() function executes on its own daemon thread and needs to ...
# - Insulate itself from shared data updates, e.g loops must copy the iterator
# - All shared data updates must be posted to the Actor main thread
#
# To Do
# ~~~~~
# * ProcessManager as-a LifeCycleManager as-a Category
#   * Design and use HyperSpace API and Storage SPI (file, SQL, MQTT ?)
#
# * CLI commands: CRUD Services in HyperSpace/Storage filesystem structure
#   - Process LifeCycle: create, enable, start, status, stop, disable, destroy
#   - Disambiguate CLI create() versus start() and destroy() versus stop() !
#
# * Aiko Dashboard plug-in support: Process CRUD et al plus ThreadManager view
#   * Use ThreadManager: Investigate https://github.com/DedInc/pythread
#   * TUI ASCII graphs for metrics and events
#
# * Send "process events" to Dashboard, Agents, Time-Series DataBase --> Graphs
#   - Use Open Telemetry schema
#
# * Handle process standard output and standard error
# * Relaunch failed processes, but don't retry continuously failing processes
# * Auto-scaling workers up/down: same host, distributed ?
#
# * Implement process "owner" field and populate automatically
#   * Implement "destroy --force" flag to destroy other owner's processes
#   * System processes owned by "aiko", e.g ProcessManager, MQTT, Registrar
#
# * Consider only one primary ProcessManager per host (refactor Registrar code)
#   * Don't allow two ProcessManagers with the same name
# * ProcessManager primary / watchdog: monitor and relaunch primary
# * Unify ProcessManagers on different hosts in the same namespace ?
#     * All the home/office servers plus handle mobile laptops (discovery) ?
#
# - aiko_process run --exit_no_processes -enp  # No running processes, then exit
#
# - Track runtime (seconds since started) for ProcessManager
#   - All process metrics: owner, restarts, runtime and resource usage (psutil)
#
# * Test on Windows operating system
# * Support Operating System specific: systemd, launchd and task scheduler
#
# - Shell pipe: (DataSource | shell command 0 | shell command 1 | DataTarget)
#   - DataSource file:// with FileSystemEventPatternMatch() for any MediaType
#
# - Consider process specific "process_exit_handler()" implementations ?
#
# - Consider using "selectors" or "asyncio.create_subprocess_exec()"
# - Check out https://pypi.org/project/python-daemon
# - Use of threading versus multiprocessing
#   - https://docs.python.org/2/library/multiprocessing.html
#   - https://hackernoon.com/concurrent-programming-in-python-is-not-what-you-think-it-is-b6439c3f3e6a

from abc import abstractmethod
import click
import importlib
import json
import os
from subprocess import Popen
import sys
from threading import Thread
import time

from aiko_services.main import *
from aiko_services.main.utilities import *

__all__ = ["ProcessManager"]

ACTOR_TYPE = "process_manager"
VERSION = 0
PROTOCOL = f"{SERVICE_PROTOCOL_AIKO}/{ACTOR_TYPE}:{VERSION}"

_GRACE_TIME=5.0  # seconds
_MY_UID="000000"  # self-aware 🤔
_PROCESS_POLL_TIME = 1.0  # seconds
_RESPONSE_TOPIC = f"{aiko.topic_in}"
_THREAD_NAME=f"aiko-{ACTOR_TYPE}"

# --------------------------------------------------------------------------- #

class ProcessManager(Actor):
    Interface.default("ProcessManager",
        "aiko_services.main.process_manager.ProcessManagerImpl")

    @abstractmethod
    def create(self, command, arguments=None, uid=None):
        pass

    @abstractmethod
    def destroy(self, uid, kill=False):
        pass

    @abstractmethod
    def dump(self):
        pass

    @abstractmethod
    def exit(self, grace_time=_GRACE_TIME):
        pass

    @abstractmethod
    def list(self, topic_path_response, uid=None):
        pass

# --------------------------------------------------------------------------- #

class ProcessCurrent:
    def __init__(self):
        self.pid = os.getpid()
        self.stdin = sys.stdin
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        self.returncode = None  # Current process is still running 😂

    def poll(self):
        return self.returncode

    def wait(self, timeout=None):
        raise RuntimeError("Cannot wait on the current process")

class ProcessManagerImpl(ProcessManager):
    def __init__(self, context, definition_pathname, watchdog,
        process_exit_handler=None):

        context.get_implementation("Actor").__init__(self, context)

        if process_exit_handler:
            self.process_exit_handler=process_exit_handler
        else:
            self.process_exit_handler=self._process_exit_handler_default
        self.processes = {
            _MY_UID: {  # self-aware 🤔
                "command_line": ["aiko_process", "run"],
                "process": ProcessCurrent()
            }
        }
        self.uid_default = 1

        self.share = {
            "lifecycle": "ready",
            "log_level": get_log_level_name(self.logger),
            "source_file": f"v{VERSION}⇒ {__file__}",
            "definition_pathname": definition_pathname,
            "metrics": {
                "created": 1,  # Including self, i.e this ProcessManager 😅
                "running": len(self.processes),
            #   "runtime": 0                                    # TODO: Runtime
            },
            "watchdog": watchdog
        }
        self.ec_producer = ECProducer(self, self.share)
        self.ec_producer.add_handler(self._ec_producer_change_handler)

        self.thread = Thread(target=self._run, daemon=True, name=_THREAD_NAME)
        self.thread.start()                           # TODO: Use ThreadManager

        self._parse_definition(definition_pathname)

    def __str__(self):
        result = ""
        for uid in self.processes.keys():
            uid_str = self._get_uid_str(uid)
            new_line = "\n" if result else ""
            result += f"{new_line}  {uid_str}"
        return result

    def create(self, command, arguments=None, uid=None):
        if not uid:
            uid = f"{self.uid_default:06d}"
            self.uid_default += 1
        if uid in self.processes:
            self.logger.warning(
                f"Create: <{uid}> exists, process not started: {command}")
            return

        command_line = [command]
        file_extension = os.path.splitext(command)[-1]
        if file_extension not in [".py", ".sh"]:     # TODO: Other extensions ?
            specification = None
            try:
                importlib.util.find_spec
            except AttributeError:  # Python < 3.4
                specification = importlib.find_loader(command)  # pylint: disable=deprecated-method
            else:
                specification = importlib.util.find_spec(command)
            if specification:
                command_line = [specification.origin]
                self.logger.info(f'Module path resolved to: "{command_line}"')

        if arguments:
            command_line.extend(arguments)
        self.logger.info(f"Create: <{uid}> {command_line}")
        try:
            process = Popen(command_line, bufsize=0, shell=False)
            self.processes[uid] = {
                "command_line": command_line,
                "process": process
            }
            self.ec_producer.update("metrics.running", len(self.processes))
            self.ec_producer.update("metrics.created",
                self.share["metrics"]["created"] + 1)
        except FileNotFoundError as file_not_found_error:
            self.logger.warning(f"{file_not_found_error}")
        except Exception as exception:
            name = exception.__class__.__name__
            self.logger.error(f"Unexpected exception: {name}: {exception}")

    def destroy(self, uid, kill=False):
        if uid != _MY_UID and uid in self.processes:
            if self._is_running(uid):
                process = self.processes[uid]["process"]
                if kill:
                    process.kill()       # SIGKILL: Can't catch, no clean-up
                else:
                    process.terminate()  # SIGTERM: Can catch and clean-up

    def dump(self):
        if len(self.processes):
            state = f" ...\n{self}"
        else:
            state = ": no processes"
        self.logger.info(f"Dump state{state}")

    def _ec_producer_change_handler(self, command, item_name, item_value):
        if item_name == "log_level":
            self.logger.setLevel(str(item_value).upper())

    def exit(self, grace_time):
        grace_time = parse_int(grace_time, _GRACE_TIME)
        for uid in self.processes.keys():
            self.destroy(uid, kill=False)  # Graceful 😇

        self.ec_producer.update("metrics.running", len(self.processes))
        time.sleep(grace_time)    # TODO: Improve how to wait for all processes

        for uid in self.processes.keys():
            self.destroy(uid, kill=True)   # Brutal ☠️
        aiko.process.terminate()

    def _get_process_info(self, uid):
        command = pid = "unknown"
        if uid in self.processes:
            process_info = self.processes[uid]
            command = process_info["command_line"]
            pid = process_info["process"].pid
        return command, pid

    def _get_return_code(self, uid):
        return_code = "uid_unknown"
        if uid in self.processes:
            return_code = self.processes[uid]["process"].poll()
        return return_code

    def _get_uid_str(self, uid):
        command, pid = self._get_process_info(uid)
        return f"<{uid}> PID: {pid} {command[0]}"

    def _is_running(self, uid):
        return self._get_return_code(uid) is None

    def list(self, topic_path_response, uid=None):
        responses = []
        for uid_key in self.processes.keys():
            if uid is None or uid == uid_key:
                command, pid = self._get_process_info(uid_key)
                response = f"{uid_key} {pid} {' '.join(command)}"
                responses.append(response)
        aiko.message.publish(
            topic_path_response, f"(item_count {len(responses)})")
        for response in responses:
            aiko.message.publish(
                topic_path_response, f"(response {response})")

    def _parse_definition(self, definition_pathname):
        diagnostic = None
        if definition_pathname:
            if definition_pathname.endswith(".json"):
                try:
                    commands = json.load(open(definition_pathname, "r"))
                    for command in commands:
                        command = command.split()
                        self.create(command[0], command[1:])
                except json.decoder.JSONDecodeError as json_decode_error:
                    diagnostic = f"Definition format: {json_decode_error}"
            else:
                diagnostic = 'Definition must be a ".json" file: ' +  \
                    definition_pathname
        if diagnostic:
            raise SystemExit(f"Error: {diagnostic}")

    def _process_exit_handler_default(self, uid):
        if uid in self.processes:
            uid_str = self._get_uid_str(uid)
            return_code = self._get_return_code(uid)
            process_info = f"{uid_str}, return_code: {return_code}"
            self.logger.info(f"Exit {process_info}")

    def _remove(self, uids):
        for uid in uids:
            if uid in self.processes:
                if self.process_exit_handler:
                    self.process_exit_handler(uid)
                del self.processes[uid]
        self.ec_producer.update("metrics.running", len(self.processes))

############################################################################
# _run() executes on a daemon thread, see concurrency design notes (above) #
############################################################################

    def _run(self):
        while True:
            zombies = []
            for uid in list(self.processes.keys()):  # safe
                if not self._is_running(uid):
                    zombies.append(uid)
            if zombies:
                self._post_message(ActorTopic.IN, "_remove", [zombies])  # safe
            time.sleep(_PROCESS_POLL_TIME)

# --------------------------------------------------------------------------- #

def get_service_filter(name=None, owner="*"):
    name = name if name else get_hostname()
    return ServiceFilter("*", name, PROTOCOL, "*", owner, "*")

@click.group()

def main():
    """Create, Read/List, Update and Destroy Processes"""
    pass

@main.command(name="create",
    context_settings=dict(allow_interspersed_args=False), no_args_is_help=True)
@click.option("--name", "-n", type=str, default=None,
    help="ProcessManager name, default is the local hostname")
@click.option("--uid", "-u", type=str, default=None,
    help="Unique name for referring to the process")
@click.argument("command", type=str)
@click.argument("arguments", nargs=-1, type=click.UNPROCESSED)

def create_command(name, uid, command, arguments):
    """Create Process

    aiko_process create [--name NAME] [--uid UID] COMMAND [ARGUMENTS ...]

    \b
    • NAME:       ProcessManager name, default is the local hostname
    • UID:        Unique IDentifier
    • COMMAND:    Command name
    • ARGUMENTS:  Command arguments
    """

    do_command(ProcessManager, get_service_filter(name),
        lambda actor: actor.create(command, arguments, uid), terminate=True)
    aiko.process.run()

@main.command(name="destroy", no_args_is_help=True)
@click.option("--kill", "-k", is_flag=True,
    help="Use SIGKILL, process can't catch and no clean-up")
@click.option("--name", "-n", type=str, default=None,
    help="ProcessManager name, default is the local hostname")
@click.argument("uid", type=str)

def destroy_command(kill, name, uid):
    """Destroy Process

    aiko_process destroy [--kill] [--name NAME] UID

    \b
    • NAME: ProcessManager name, default is the local hostname
    • UID:  Unique IDentifier
    """

    do_command(ProcessManager, get_service_filter(name),
        lambda actor: actor.destroy(uid, kill), terminate=True)
    aiko.process.run()

@main.command(name="dump", help="Dump ProcessManager state")
@click.option("--name", "-n", type=str, default=None,
    help="ProcessManager name, default is the local hostname")

def dump_command(name):
    do_command(ProcessManager, get_service_filter(name),
        lambda actor: actor.dump(), terminate=True)
    aiko.process.run()

@main.command(name="exit", help="Exit ProcessManager")
@click.option("--grace_time", "-gt", type=int, default=_GRACE_TIME,
    help="Wait time before killing child processes")
@click.option("--name", "-n", type=str, default=None,
    help="ProcessManager name, default is the local hostname")

def exit_command(grace_time, name):
    do_command(ProcessManager, get_service_filter(name),
        lambda actor: actor.exit(grace_time), terminate=True)
    aiko.process.run()

@main.command(name="list")
@click.option("--name", "-n", type=str, default=None,
    help="ProcessManager name, default is the local hostname")
@click.argument("uid", type=str, required=False, default=None)

def list_command(name, uid):
    """List Processes

    aiko_process list [--name NAME] [UID]

    \b
    • NAME: ProcessManager name, default is the local hostname
    • UID:  Unique IDentifier to match
    """

    def response_handler(response):
        if len(response):
            output = f"Processes ..."
            for process_record in response:
                uid = process_record[0]
                pid = process_record[1]
                command = process_record[2:]
                output += f"\n  uid: {uid}, pid: {pid}, {' '.join(command)}"
        else:
            output = "No processes"
        print(output)

    do_request(ProcessManager, get_service_filter(name),
        lambda actor: actor.list(_RESPONSE_TOPIC, uid),
        response_handler, _RESPONSE_TOPIC, terminate=True)
    aiko.process.run()

@main.command(name="run")
@click.option("--name", "-n", type=str, default=None,
    help="ProcessManager name, default is the local hostname")
@click.option("--watchdog", "-w", is_flag=True,
    help="Monitor ProcessManager, if required relauch it")
@click.argument("definition_pathname", required=False, default=None)

def run_command(name, watchdog, definition_pathname):
    """Run ProcessManager

    aiko_process run [--name NAME] [--watchdog] [HYPERSPACE_PATHNAME]

    \b
    • NAME:                ProcessManager name, default is the local hostname
    • HYPERSPACE_PATHNAME: HyperSpace storage file-system location
    """

    name = name if name else get_hostname()

    tags = ["ec=true"]       # TODO: Add ECProducer tag before add to Registrar
    init_args = actor_args(name, None, None, PROTOCOL, tags)
    init_args["definition_pathname"] = definition_pathname
    init_args["watchdog"] = watchdog
    process_manager = compose_instance(ProcessManagerImpl, init_args)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



import io
import json
import sys
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union
from uuid import UUID

import numpy as np
from PIL import Image

from highlighter.client import HLJSONEncoder
from highlighter.client.base_models.entity import Entity
from highlighter.core import LabeledUUID
from highlighter.core.data_models.data_sample import DataSample

from .base_capability import (
    Capability,
    ContextPipelineElement,
    DataSourceType,
    StreamEvent,
)

__all__ = [
    "EntityWriteFile",
    "WriteStdOut",
    "ImageWriteStdOut",
    "ImageWrite",
    "TextToStdout",
    "Print",
]

# ToDo: The attribute(s) returned by an llm or any model
#       will need to be configured at run time. The the exception
#       of a hand-full of attributes (pixel-location, object-class, ...)
#       we can't know ahead of time what attribute the output of
#       the model will represent.
TEXT_ATTRIBUTE_UUID = LabeledUUID(int=2, label="response")


class BaseEntityWrite(Capability):

    def __init__(self, context: ContextPipelineElement):
        super().__init__(context)
        self.frame_entities = dict()

    def get_task_id(self, stream) -> str:
        return stream.variables.get("task_id", None)

    def _get_source_file_location(self, stream):
        # ToDo: Find a better palce to put/get this from
        #       see also, ImageDataSource.process_frame
        source_info = stream.variables.get("source_info", {})
        source_file_location = source_info.get("source_file_location", None)
        if source_file_location is not None:
            return Path(source_file_location)
        return None

    def on_per_frame(self, stream, entities, data_sample):
        pass

    def process_frame(
        self, stream, data_samples: List[DataSample], entities: Dict[UUID, Entity]
    ) -> Tuple[StreamEvent, dict]:
        for data_sample in data_samples:
            data_sample_entities = {
                e.id: e
                for e in entities.values()
                if len(e.annotations) == 0 or e.annotations[0].data_file_id == data_sample.data_file_id
            }
            self.on_per_frame(stream, data_sample_entities, data_sample)
        self.frame_entities[stream.frame_id] = entities
        return StreamEvent.OKAY, {}

    def on_stop_stream(self, stream, stream_id, entities):
        """Note this will not be called if you're calling `pipeline.process_frame`
        directly. Because this is called when a stream is stopped
        """
        pass

    def stop_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[Dict]]:
        self.on_stop_stream(stream, stream_id, self.frame_entities)
        return StreamEvent.OKAY, {}


class EntityWriteFile(BaseEntityWrite):

    class StreamParameters(BaseEntityWrite.StreamParameters):
        """Can contain the following placeholders:

            {frame_id}
            {task_id}
            {timestamp}

        for example:
            per_frame_output_file = 'output_{frame_id}_{timestamp}.json'
        """

        per_frame_output_file: Optional[str] = None

        """Can contain the following placeholders:

            {task_id}
            {timestamp}
        """
        stop_stream_output_file: Optional[str] = None

    def _timestamp(self):
        return datetime.now().strftime("%Y%m%d%H%M%S%f")

    def get_per_frame_output_file_path(self, stream, data_sample):
        task_id = stream.stream_id
        frame_id = data_sample.media_frame_index

        return self.stream_parameters(stream.stream_id).per_frame_output_file.format(
            frame_id=frame_id,
            task_id=task_id,
            timestamp=self._timestamp(),
        )

    def on_per_frame(self, stream, entities, data_sample):
        if self.stream_parameters(stream.stream_id).per_frame_output_file:
            output_file_path = self.get_per_frame_output_file_path(stream, data_sample)
            Path(output_file_path).parent.mkdir(exist_ok=True, parents=True)
            output_str = json.dumps(entities, indent=2, sort_keys=True, cls=HLJSONEncoder)
            with open(output_file_path, "w") as f:
                f.write(output_str)
                self.logger.debug(f"{self.my_id()}: wrote {len(entities)} entities to {output_file_path} ")

    def get_on_stop_stream_output_file_path(self, stream):
        task_id = stream.variables.get("task_id", None)
        return self.stream_parameters(stream.stream_id).stop_stream_output_file.format(
            stream_id=stream.stream_id,
            task_id=task_id,
        )

    def on_stop_stream(self, stream, stream_id, all_entities):
        stop_stream_output_file = self.stream_parameters(stream_id).stop_stream_output_file
        if stop_stream_output_file:
            self.logger.debug(f"Writing stop_stream_output_file: {stop_stream_output_file}")
            output_file_path = self.get_on_stop_stream_output_file_path(stream)
            Path(output_file_path).parent.mkdir(parents=True, exist_ok=True)
            with open(output_file_path, "w") as f:
                f.write(json.dumps(all_entities, indent=2, sort_keys=True, cls=HLJSONEncoder))


class WriteStdOut(BaseEntityWrite):

    def on_per_frame(self, stream, entities, data_sample):

        output = {"frame_data": entities, "frame_id": stream.frame_id, "data_file": data_sample.data_file_id}
        print(json.dumps(output, cls=HLJSONEncoder, indent=2), file=sys.stdout)


class ImageWriteStdOut(Capability):

    def process_frame(self, stream, data_samples: List[DataSourceType]) -> Tuple[StreamEvent, Optional[Dict]]:
        image = data_samples[0].content
        output_buffer = io.BytesIO()
        if isinstance(image, np.ndarray):
            Image.fromarray(image).save(output_buffer, format="PNG")
        elif isinstance(image, Image.Image):
            image.save(output_buffer, format="PNG")

        sys.stdout.buffer.write(output_buffer.getvalue())

        return StreamEvent.OKAY, {}


class ImageWrite(Capability):

    class StreamParameters(Capability.StreamParameters):
        output_dir: str

        # Can use placeholders, {file_id}, {media_frame_index}, {original_source_url}
        output_pattern: str

    def process_frame(self, stream, data_samples: List[DataSample]) -> Tuple[StreamEvent, dict]:
        output_dir = Path(self.stream_parameters(stream.stream_id).output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)

        for ds in data_samples:
            filename = self.stream_parameters(stream.stream_id).output_pattern.format(
                file_id=ds.data_file_id,
                media_frame_index=ds.media_frame_index,
                original_source_url=ds.original_source_url,
            )
            dest = output_dir / filename
            image = ds.content
            if isinstance(image, np.ndarray):
                Image.fromarray(image).save(dest)
            elif isinstance(image, Image.Image):
                image.save(dest)

        return StreamEvent.OKAY, {}


class TextToStdout(Capability):
    def process_frame(self, stream, text: str) -> Tuple[StreamEvent, Union[Dict, str]]:
        sys.stdout.write(text)
        # sys.stdout.write(chr(28))  # ASCII 28 (Record Separator)
        sys.stdout.flush()
        return StreamEvent.OKAY, {}


class Print(Capability):
    class StreamParameters(Capability.StreamParameters):
        prefix: str = ""

    def process_frame(self, stream, **kwargs) -> Tuple[StreamEvent, Union[Dict, str]]:
        string = self.stream_parameters(stream.stream_id).prefix + str(kwargs)
        print(string)
        return StreamEvent.OKAY, {}


class FrameDataWriteJson(Capability):
    class StreamParameters(Capability.StreamParameters):
        output_folder: str

    def __init__(self, context):
        super().__init__(context)
        self.frame_data = {}

    def start_stream(self, stream, stream_id, use_create_frame=True):
        super().start_stream(stream, stream_id, use_create_frame=use_create_frame)
        self.frame_data[stream_id] = {}
        return StreamEvent.OKAY, {}

    def stop_stream(self, stream, stream_id):
        super().stop_stream(stream, stream_id)
        with open(Path(self.stream_parameters(stream_id).output_folder) / f"{stream_id}.json", "w") as f:
            f.write(json.dumps(self.frame_data[stream_id], cls=HLJSONEncoder, indent=2, sort_keys=True))
        del self.frame_data[stream_id]
        return StreamEvent.OKAY, None

    def process_frame(self, stream, **kwargs):
        self.frame_data[stream.stream_id][stream.frame_id] = kwargs
        return StreamEvent.OKAY, {}



import json
import logging
import sys
from typing import List
from urllib.parse import urlparse

import click

from highlighter.client import CloudAgent
from highlighter.client.agents import create_agent_token, create_machine_agent_version
from highlighter.client.gql_client import HLClient
from highlighter.client.json_tools import HLJSONEncoder
from highlighter.core.config import HighlighterRuntimeConfigError
from highlighter.core.runtime import Runtime

logger = logging.getLogger(__name__)


@click.group("agent")
@click.pass_context
def agent_group(ctx):
    pass


@agent_group.command("start")
@click.option("--task-ids", "-t", type=str, default=None, help="comma separate for multiple")
@click.option("--step-id", "-i", type=str, default=None)
@click.option("--stream-definitions-file", type=str)
@click.option("--dump-definition", type=str, default=None)
@click.option("--allow-non-machine-user", is_flag=True, default=False)
@click.option("--server", is_flag=True, default=False)
@click.argument("agent_definition", type=click.Path(dir_okay=False, exists=False))
@click.argument("inputs", nargs=-1, type=click.STRING, required=False)
@click.pass_context
def _start(
    ctx,
    task_ids,
    step_id,
    stream_definitions_file,
    dump_definition,
    allow_non_machine_user,
    server,
    agent_definition,
    inputs,
):
    """Start a local Highlighter Agent to process data either from your local machine or from Highlighter tasks.

    When processing URLs or local files, a single stream is created to process each URL or file.
    The Agent definition must have its first element as a
    DataSourceCapability, such as ImageDataSource, VideoDataSource,
    TextDataSource, JsonArrayDataSource, etc. The examples below assume
    the use of ImageDataSource.

    When processing Highlighter tasks, a single stream is created for each
    task. The Agent definition should use AssessmentRead as the first element in this case.
    Note: When processing tasks, use a GraphQL API key specific to the agent being run. You can
    create this using 'hl agent create-token'.

    URLs and files can be passed as arguments.
    Alterlatively, the argument '-' can be given and the contents of an image
    can be passed via stdin (when using the ImageDataSource capability) or JSON can be
    piped in (when using the JsonArrayDataSource capability).

    If '--server' is specified, the runtime won't shutdown after processing the specified
    inputs.

    Examples:

      \b
      1. Start an agent against a single image path
      \b
        > hl agent start agent-def.json images/123.jpg

      \b
      2. Start an agent against a multiple image paths
      \b
        > hl agent start agent-def.json $(find images/ -name *.jpg)
        OR
        > find images/ -name *.jpg -print0 | xargs -0 hl agent start agent-def.json

      \b
      3. Cat the contents of an image to an agent
      \b
        > cat images/123.jpg | hl agent start agent-def.json -

      \b
      4. Pipe json data to stdin
      \b
        > cat '[{"foo": "bar"},{"foo": "baz"}]' | hl agent start agent-def.json -

      \b
      5. Process streams defined in a file
      \b
        > hl agent start agent-def.json --stream-definitions-file streams.json

      \b
      6. Process tasks from a Highlighter machine-step, using a local agent definition
      \b
        > STEP_UUID=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        > hl agent start agent-def.json --step-id "$STEP_UUID"

      \b
      7. Process tasks from a Highlighter machine-step, using an agent definition from Highlighter
      \b
        > STEP_UUID=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        > AGENT_UUID=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        > hl agent start "$AGENT_UUID" --step-id "$STEP_UUID"

    """
    try:
        stream_definitions = []
        files = []
        urls = []
        if len(inputs) > 0:
            if inputs[0] == "-":
                urls.append("hlpipe://")
            else:
                for input_arg in inputs:
                    if urlparse(input_arg).scheme == "":
                        files.append(input_arg)
                    else:
                        urls.append(input_arg)
        if stream_definitions_file is not None:
            with open(stream_definitions_file) as sdf:
                stream_definitions = json.load(sdf)
        runtime = Runtime(
            agent_definition=agent_definition,
            dump_definition=dump_definition,
            allow_non_machine_user=allow_non_machine_user,
            hl_cfg=ctx.obj.get("hl_cfg"),
            hl_client=ctx.obj.get("client"),
            queue_response=ctx.obj.get("queue_response"),
        )
        runtime.run(
            stream_definitions=stream_definitions,
            files=files,
            urls=urls,
            task_ids=task_ids,
            step_id=step_id,
            server=server,
        )
    except HighlighterRuntimeConfigError as e:
        logger.error(f"Failed to load configuration: {e}")
        sys.exit(2)
    except Exception as exc:
        logger.error(f"{type(exc).__qualname__}: {exc}")
        sys.exit(1)


@agent_group.command("create-token")
@click.option("--machine-agent-version-id", type=str)
@click.option("--machine-agent-name", type=str, required=False)
@click.option("--machine-agent-version-name", type=str, required=False)
def _create_token(
    machine_agent_version_id,
    machine_agent_name,
    machine_agent_version_name,
):
    """Create an access token for an agent

    Once an access token has been created, run an agent with that identity by
    setting `HL_WEB_GRAPHQL_API_TOKEN=<new-token>` before running `hl agent start`

    Either provide the ID of the machine-agent-version in Highlighter, or
    specify a new machine-agent name and version-name to create a new machine-agent-version
    for your agent.
    """
    if machine_agent_version_id is None:
        if machine_agent_name is not None or machine_agent_version_name is not None:
            raise ValueError(
                "Must specify either 'machine_agent_token', give a machine-agent version "
                "ID as the agent definition, or specify both 'machine_agent_name' and "
                "'machine_agent_version'"
            )
        machine_agent_version = create_machine_agent_version(machine_agent_name, machine_agent_version_name)
        machine_agent_version_id = machine_agent_version.id
    machine_agent_token = create_agent_token(machine_agent_version_id)
    # Print to stdout rather than log
    print(machine_agent_token)


@agent_group.command("list")
@click.option("--cloud", is_flag=True)
def _list(cloud):
    """List running agents"""
    if not cloud:
        raise NotImplementedError(
            "Can currently only list agents running in the cloud with the '--cloud' flag"
        )
    try:
        client = HLClient.get_client()
        cloud_agents = client.cloudAgents(return_type=List[CloudAgent])
        print(json.dumps(cloud_agents, indent=4, cls=HLJSONEncoder))
    except Exception as e:
        print(f"Error listing agents. {type(e).__qualname__}: {e}")



import io
import logging
import os
from dataclasses import dataclass
from fractions import Fraction
from pathlib import Path
from typing import BinaryIO, Iterable, Union

import av
import numpy as np
from PIL import Image

from highlighter.agent.utilities import EntityAggregator, FileAvroEntityWriter
from highlighter.client import ENTITY_AVRO_SCHEMA
from highlighter.client.base_models.entities import Entities
from highlighter.core.data_models.data_sample import DataSample
from highlighter.io.base import PayloadWriter

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class EncodeSettings:
    codec: str = "h264"
    pix_fmt: str = "yuv420p"
    crf: int = 23
    preset: str = "medium"
    scenecut: str = "0"


class ImageWriter(PayloadWriter):
    """
    Serialise one or many image samples as a single **PNG**.

    * If exactly one sample is supplied, the payload is just that image.
    * If more than one sample is supplied, the images are tiled into a
      square-ish grid:  ``cols = ceil(sqrt(N)), rows = ceil(N/cols)``.
    """

    def __init__(self, *, mode: str = "RGB", extension: str = "PNG"):
        """
        Parameters
        ----------
        mode
            PIL/Pillow image mode for the final canvas (defaults to ``"RGB"``).
        extension
            PIL/Pillow image extension for the final canvas. Required if `write`ing
            to a `sink` of type `BinaryIO`. If `sink` is a `str` or `os.PathLike`
            then `extension is ignored. (defaults to ``"PNG"``).


        """
        self.mode = mode
        self.extension = extension

    # The signature mirrors VideoWriter so DataFile.save_local can treat every
    # writer uniformly.
    def write(
        self,
        samples: Iterable[DataSample],
        sink: Union[str, os.PathLike, BinaryIO],
    ) -> None:
        samples = list(samples)
        if not samples:
            raise ValueError("ImageWriter.write() received no samples")

        # Convert every DataSample → PIL.Image
        pil_images = []
        for s in samples:
            if isinstance(s.content, Image.Image):
                img = s.content.convert(self.mode)
            else:
                arr = s.to_ndarray()
                if arr.ndim == 2:  # greyscale → RGB
                    arr = np.stack([arr] * 3, axis=-1)
                img = Image.fromarray(arr).convert(self.mode)
            pil_images.append(img)

        # Tile if necessary
        if len(pil_images) == 1:
            final_img = pil_images[0]
        else:
            w, h = pil_images[0].size
            n = len(pil_images)
            cols = math.ceil(math.sqrt(n))
            rows = math.ceil(n / cols)
            canvas = Image.new(self.mode, (cols * w, rows * h))
            for idx, img in enumerate(pil_images):
                r, c = divmod(idx, cols)
                canvas.paste(img.resize((w, h)), (c * w, r * h))
            final_img = canvas

        # Write to the requested sink
        if isinstance(sink, (str, Path)):
            with open(sink, "wb") as f:
                final_img.save(f)
        else:
            final_img.save(sink, format=self.extension)


class TextWriter(PayloadWriter):
    def write():
        pass


class EntityWriter(PayloadWriter):
    def __init__(self, *, extension: str = "avro"):
        """
        Parameters
        ----------
        extension
            file extension which defaults to avro
        """
        self.extension = extension

    def write(
        self,
        samples: Iterable[DataSample],
        sink: Union[str, os.PathLike, BinaryIO],
    ) -> None:
        """
        Encode all Entities into *sink*
        """
        writer = FileAvroEntityWriter(
            ENTITY_AVRO_SCHEMA,
            sink,
        )
        agg = EntityAggregator(
            minimum_track_frame_length=3,
            minimum_embedding_in_track_frame_length=3,
            writer=writer,
        )
        samples = list(samples)
        for sample in samples:
            if not isinstance(sample.content, Entities):
                raise ValueError("EntityWriter requires Entities sample content")
            agg.append_entities(list(sample.content))
        agg.write()


class VideoWriter(PayloadWriter):
    """
    Stream-encode frames into a video sink with efficient, C-based resizing:
      • configurable frame_rate, bit_rate, resolution
      • pre-flight probing for resolution and bitrate
      • `av.VideoFrame.reformat` for on-the-fly scaling (fast C)
    """

    supports_resize = True

    def __init__(
        self,
        *,
        frame_rate: float = 24.0,
        bit_rate: int | None = None,
        resolution: tuple[int, int] | None = None,
        extension: str = "mp4",
        settings: EncodeSettings = EncodeSettings(),
    ):
        self.frame_rate = frame_rate
        self.bit_rate = bit_rate
        self.resolution = resolution
        self.settings = settings
        self.extension = extension

    def _prepare(self, first_frame: DataSample) -> tuple[int, int]:
        """
        Determine target width/height and estimate bit_rate if unset.
        """
        arr = first_frame.to_ndarray()
        h, w = arr.shape[:2]
        if self.resolution:
            width, height = self.resolution
        else:
            width, height = w, h
        if self.bit_rate is None:
            raw_bytes = arr.nbytes
            # use 25% of raw size as heuristic
            self.bit_rate = int(raw_bytes * self.frame_rate * 8 * 0.25)
        return width, height

    def write(
        self,
        samples: Iterable[DataSample],
        sink: Union[str, os.PathLike, BinaryIO],
    ) -> None:
        """
        Encode all frames into *sink*, streaming scaled frames via C.
        """
        frames = list(samples)
        if not frames:
            raise ValueError("VideoWriter.write() received no samples")

        width, height = self._prepare(frames[0])

        close_sink = False
        if isinstance(sink, (str, Path)):
            sink = open(sink, "wb")  # type: ignore[assignment]
            close_sink = True

        container = av.open(sink, mode="w", format=self.extension)

        # Set metadata from first frame's recorded_at if available
        if hasattr(frames[0], "recorded_at") and frames[0].recorded_at is not None:
            first_frame_time = frames[0].recorded_at
            # ISO standard creation time (container level)
            container.metadata["creation_time"] = first_frame_time.isoformat()
            # QuickTime standard creation date
            container.metadata["©day"] = first_frame_time.isoformat()
            # Custom field for application-specific queries
            container.metadata["recorded_at"] = first_frame_time.isoformat()

        stream = container.add_stream(self.settings.codec, rate=Fraction(self.frame_rate))
        stream.width, stream.height = width, height
        stream.pix_fmt = self.settings.pix_fmt
        stream.options.update(
            {
                "crf": str(self.settings.crf),
                "preset": self.settings.preset,
                "scenecut": self.settings.scenecut,
            }
        )
        if self.bit_rate:
            stream.bit_rate = self.bit_rate

        # Set metadata at track level too
        if hasattr(frames[0], "recorded_at") and frames[0].recorded_at is not None:
            stream.metadata["creation_time"] = frames[0].recorded_at.isoformat()
            stream.metadata["recorded_at"] = frames[0].recorded_at.isoformat()
            stream.metadata["©day"] = frames[0].recorded_at.isoformat()

        for frame in frames:
            arr = frame.to_ndarray()
            if not isinstance(arr, np.ndarray):
                raise TypeError(f"Expected a NumPy ndarray, got {type(arr).__name__}")
            vf = av.VideoFrame.from_ndarray(arr, format="rgb24")
            # fast, C-level reformat for scaling/pad
            if (vf.width, vf.height) != (width, height):
                vf = vf.reformat(width=width, height=height, format="rgb24")  # C-based scaling
            for packet in stream.encode(vf):
                container.mux(packet)

        # flush and close
        for packet in stream.encode():
            container.mux(packet)
        container.close()
        if close_sink:
            sink.close()



# Aiko Services: Dependency
# ~~~~~~~~~~~~~~~~~~~~~~~~~
# Is a "reference" to a distributed Service, containing ...
# - Local or remote Service instance reference (when discovered)
# - ServiceFilter for Service discovery
# - LifeCycleManager URL for the Actor that manages the Service (load / unload)
# - Storage URL for the Actor that persists the Service Definition / Content
#
# Usage
# ~~~~~
#   topic_path = "*"
#   service_name = name
#   protocol = "*"
#   transport = "*"
#   owner = "*"
#   tags = "*"
#
#   service_filter = ServiceFilter(         # Used for Service discovery
#       topic_path, service_name, protocol, transport, owner, tags)
#   lifecycle_manager_url = None            # TODO: Implement lifecycle_manager
#   storage_url = None                      # TODO: Implement storage
#
#   init_args = dependency_args(
#       None, service_filter, lifecycle_manager_url, storage_url)
#   dependency = compose_instance(DependencyImpl, init_args)
#
# To Do
# ~~~~~
# * Consider Dependency.update() same as Category.update() for "service_filter"
#
# * Consider use of ServiceFields rather than ServiceFilter ?
#
# * Replace current Pipeline prototype implementation with this improved design
#   * Once full Dependency resolution, information flow, etc is completed
#
# - Properties: Dependency direction, Information direction, Strong, Weak, Group
#
# - DependencyManager maintaining in-memory list of Dependencies ?
#
# - Security model design and implementation

from abc import abstractmethod

__all__ = ["Dependency", "dependency_args", "DependencyImpl"]

from aiko_services.main import *

# --------------------------------------------------------------------------- #

class Dependency(Interface):
    Interface.default(
        "Dependency", "aiko_services.main.dependency.DependencyImpl")

    @abstractmethod
    def is_type(self, type_name):
        pass

    @abstractmethod
    def update(self, entry_name, service=None,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):
        pass

def dependency_args(service=None,
    service_filter=None, lifecycle_manager_url=None, storage_url=None):

    init_args = context_args()
    init_args.update({
        "service": service,
        "service_filter": service_filter,
        "lifecycle_manager_url": lifecycle_manager_url,
        "storage_url": storage_url})
    return init_args

# --------------------------------------------------------------------------- #
# "self.service" reference values ...
#     - None: Service not discovered or absent
#     - Local instance reference (running uithin the same process)
#     - Remote reference (running in a different process or host)
#
# TODO: "self.lifecycle_manager_url" and "self.storage_url" values ...
#     - None: Service (dependency) is manually started
#     - "*":  Discover ProcessManager and/or Storage (default)
#     - "URL" Specified ProcessManager and/or Storage

class DependencyImpl(Dependency):
    def __init__(self, context, service=None,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):

        self.service = service                     # Service instance reference
        self.service_filter = service_filter       # Used for discovery
        self.lifecycle_manager_url = lifecycle_manager_url
        self.storage_url = storage_url

    def is_type(self, type_name):
        return type_name.lower() == "dependency"

    def __repr__(self):
        return f"({self.service_filter} "  \
               f"{self.lifecycle_manager_url} "  \
               f"{self.storage_url})"

# Note: "entry_name" only needed for Category, not Dependency, ignore it here

    def update(self, entry_name, service=None,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):

        if service:
            self.service = service
        if service_filter:
            self.service_filter = service_filter
        if lifecycle_manager_url:
            self.lifecycle_manager_url = lifecycle_manager_url
        if storage_url:
            self.storage_url = storage_url

# --------------------------------------------------------------------------- #



import json
import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import click
from cookiecutter.generate import generate_context
from cookiecutter.main import cookiecutter
from cookiecutter.prompt import prompt_for_config

from highlighter.cli.logging import ColourStr
from highlighter.client import TrainingConfigType
from highlighter.client.gql_client import HLClient

# register custom filters for cookiecutter
from highlighter.templates.filters import *
from highlighter.trainers import _scaffold


def _do_cookiecutter(tmpdir, template_dir, context_dict):
    dst_dir = Path(tmpdir) / "template"
    shutil.copytree(template_dir, dst_dir)

    cookiecutter_json = dst_dir / "cookiecutter.json"
    with cookiecutter_json.open("w") as f:
        json.dump(context_dict, f)

    # Load the context from the Cookiecutter template
    context = generate_context(context_file=str(cookiecutter_json))
    # Prompt user for inputs
    completed_context = prompt_for_config(context)
    # Render the template
    final_path = cookiecutter(
        str(dst_dir),
        no_input=True,
        extra_context=completed_context,
        output_dir=tmpdir,
    )
    return final_path, completed_context


@click.group("generate")
@click.pass_context
def generate_group(ctx):
    pass


@generate_group.command("agent")
@click.argument(
    "agents_dir", type=click.Path(dir_okay=True, file_okay=False), required=False, default="./agents"
)
@click.pass_context
def agent_generate(ctx, agents_dir):
    """Follow the prompts to create a basic agent definition in the specified
    location.

    \b
    Usage:
        # Generate an agent def json in ./agents/
        hl generate agent

        \b
        # Generate an agent def json in ./foo/bar/
        # if ./foo/bar does not exist it will be created
        hl generate agent ./foo/bar

    \b
    Args:
        agents_dir: Directory to contain your agent definition. Default
        './agents', if it does not exist it will be created.
    """
    template_dir = Path(__file__).parent.parent / "templates" / "agent"

    context_dict = {
        "data_source_type": ["highlighter", "local_text", "local_video", "local_image"],
        "agent_name": "my_agent",
        "_root": "root_dir",
    }

    with TemporaryDirectory() as tmp:
        final_path, completed_context = _do_cookiecutter(tmp, template_dir, context_dict)
        agents_dir = Path(agents_dir)
        agents_dir.mkdir(exist_ok=True, parents=True)
        agent_def_file_name = f'{completed_context["agent_name"]}.json'
        gen_agent_path = Path(final_path) / agent_def_file_name
        new_agent_dest = agents_dir / agent_def_file_name
        _scaffold.safe_move(gen_agent_path, new_agent_dest)
    msg = ColourStr.green(f"Created {completed_context['agent_name']} in {new_agent_dest}")
    click.echo(f"{msg}")


@generate_group.command("capability")
@click.argument("capabilities_dir", type=click.Path(dir_okay=True, file_okay=False), required=True)
@click.pass_context
def capability_generate(ctx, capabilities_dir):
    """Follow the prompts to create the boilerplate for a basic capabilility, and
    optioanlly add it to an existing agent definition

    \b
    Usage:
        # Generate capabilility boilerplate in ./src/PACKAGE_NAME/capabilities/
        hl generate capability

        \b
        # Generate capabilility boilerplate in ./foo/bar/
        # if ./foo/bar does not exist it will be created
        hl generate capabilility ./foo/bar

    \b
    Args:
        capabilities_dir: Directory to put your capabilility. Default
        './src/PACKAGE_NAME/capabilities', if it does not exist it will be created.
    """
    template_dir = Path(__file__).parent.parent / "templates" / "capabilility"

    context_dict = {
        "capability_class_name": "MyCapability",
        "capability_module_name": "{{ cookiecutter.capability_class_name | camel_to_snake }}",
        "capability_description": "Does something cool",
        "_root": "root_dir",
        "_extensions": ["highlighter.templates.filters.camel_to_snake"],
    }

    with TemporaryDirectory() as tmp:
        final_path, completed_context = _do_cookiecutter(tmp, template_dir, context_dict)
        capability_py = f"{completed_context['capability_module_name']}.py"
        gen_capability_path = Path(final_path) / capability_py
        dest_capability_path = Path(capabilities_dir) / capability_py
        _scaffold.safe_move(gen_capability_path, dest_capability_path)
    msg = ColourStr.green(f"Created {completed_context['capability_class_name']} in {dest_capability_path}")
    click.echo(f"{msg}")


@generate_group.command("training-run")
@click.argument("training_run_id", type=int)
@click.argument("trainer", type=click.Choice(list(_scaffold.TrainerType)))
@click.argument(
    "ml_training_dir",
    type=click.Path(dir_okay=True, file_okay=False),
    required=False,
    default="./ml_training",
)
@click.option("--page-size", type=int, default=None)
@click.pass_context
def train_generate(ctx, training_run_id, trainer, ml_training_dir, page_size):
    """Generate the boilerplate for a model training.

    \b
    Usage:
        # Generate training boilerplate for a YOLO Detector in ./ml_training/TRAINING_RUN_ID/
        hl generate training-run TRAINING_RUN_ID yolo-det

        # Generate training boilerplate for a YOLO Detector in ./foo/bar/
        # if ./foo/bar does not exist it will be created
        hl generate training-run TRAINING_RUN_ID yolo-det ./foo/bar

    \b
    Args:
        training_run_id: The id of the Highligher Training Run
        trainer: select from supported trainers
        ml_training_dir: Directory to put your training files. Default
        './ml_training/TRAINING_RUN_ID', if it does not exist it will be created.
    """

    trainer_name = trainer
    ml_training_dir = Path(ml_training_dir)

    client: HLClient = ctx.obj["client"]
    highlighter_training_config = TrainingConfigType.from_highlighter(client, training_run_id)

    training_run_dir = ml_training_dir / str(training_run_id)
    if training_run_dir.exists():
        _scaffold.ask_to_remove(training_run_dir, terminate=True)

    training_run_dir.mkdir(parents=True, exist_ok=False)

    if "yolo" in trainer_name:
        from highlighter.trainers.yolov11 import YoloV11Trainer as Trainer
    else:
        raise ValueError(f"Unable to determine trainer from '{trainer_name}'")

    page_size = page_size if page_size is not None else ctx.obj["hl_cfg"].pagination_page_size
    trainer = Trainer(training_run_dir, highlighter_training_config, trainer_name)
    trainer.generate_boilerplate()
    trainer.get_datasets(client, page_size=page_size)

    click.echo(f"{trainer_name} template generated at {training_run_dir}")
    cmd = ColourStr.green(f"hl train start {training_run_dir}")
    cfg_path = ColourStr.green(f"{trainer.config_path}")
    click.echo(f"OPTIONAL, edit the config at {cfg_path}")
    click.echo(f"Start training with: `{cmd}`")



#!/usr/bin/env python3
#
# Aiko Service: Category(Dependency)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Create, Read, Update and Destroy a Category
# Add to, List and Remove from a Category, which is a group of Entries
# Entries can be either a Category or a Dependency
#
# Usage
# ~~~~~
# ./category.py --help
# ./category.py run     CATEGORY_NAME  # Creates Category instance for testing
# ./category.py exit    CATEGORY_NAME  # Terminates running Category instance
#
# ./category.py add     CATEGORY_NAME ENTRY_NAME -p p -o a -t key_0=value_0
# ./category.py list    CATEGORY_NAME [ENTRY_NAME] [--long_format | -l]
# ./category.py remove  CATEGORY_NAME ENTRY_NAME
# ./category.py update  CATEGORY_NAME ENTRY_NAME -p protocol -t key_0=value_0
#
# To Do
# ~~~~~
# - Categories: HyperSpace, Registrar, LifeCycleManager, WorkSpace
#   - LifeCycleManagers: HyperSpace, ProcessManager, Pipeline
#   - Distributed Garbage Collection: Use of Leases by LifeCycleManagers ?
#     - LifeCycleManagers also "reference count" Categories references ?
#     - Interaction with Storage for managing persistence life-time ?
#
# - Consider LifeCycleManager as-a Category as-a DependencyManager
#
# * Implement "./category.py repl" for interactive CRUD, etc
#   - REPL command to discover then add, update or remove a running Service
#
# * Fix: Implement "__repr__()", "__str__()" and refactor "list()" to use them ?
#   - https://stackoverflow.com/questions/1436703/what-is-the-difference-between-str-and-repr
#
# * Fix: list(): "(dn (None * * * * (a=b c=d)) 0: 0:)"  # None --> 0:
#
# * Incorporate Dependency design and implementation, e.g strong, weak, group
#   * Disambiguate function delivery to the Category versus the Entries 🤔
#
# * Timestamps: Category create, Dependency add(), remove(), update()
# * Category Metrics: add(), list(), remove(), update() rates, total counts
#
# * Implement "category.py list" filters, e.g by owner, by protocol, etc
#
# - Design a CategoryManager interface for use by HyperSpace et al
#
# - Categories that span different HyperSpace instances, e.g different "roots"
#
# - Consider design and implementation of "self-changing custom" Categories
#   - Custom Category implementations may want many Categories per Process
#
# - Design and implement Queue(Category) with different persistence options
#   - Implement "queue://" DataScheme, DataSource and DataTarget
#
# * Implement Category "owner" field and populate automatically
#   * System Categories owned by "aiko"
#   * Security design and implementation: owners, roles, ACLs

from abc import abstractmethod
import click

from aiko_services.main import *

__all__ = ["Category"]

_RESPONSE_TOPIC = f"{aiko.topic_in}"

# --------------------------------------------------------------------------- #

class Category(Actor, Dependency):
    Interface.default("Category", "aiko_services.main.category.CategoryImpl")

    @abstractmethod
    def add(self, entry_name,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):
        pass

    @abstractmethod
    def exit(self):
        pass

    @abstractmethod
    def list(self, topic_path_response, entry_name=None, recursive=False):
        pass

    @abstractmethod
    def remove(self, entry_name):
        pass

# --------------------------------------------------------------------------- #

class CategoryImpl(Category):
    def __init__(self, context, service_filter=None):
        context.call_init(self, "Actor", context)
        context.call_init(self, "Dependency", context,
            service_filter=service_filter)

        self.dependency = context.get_implementation("Dependency")    # methods
        self.entries = {}                         # Categories and Dependencies

        self.share.update({                       # Inherit from Actor
            "source_file": f"v{CATEGORY_VERSION}⇒ {__file__}",
            "entries": len(self.entries)
        })
        self.ec_producer = ECProducer(self, self.share)
        self.ec_producer.add_handler(self._ec_producer_change_handler)

    def add(self, entry_name,
        service_filter, lifecycle_manager_url=None, storage_url=None):

        if service_filter and not isinstance(service_filter, ServiceFilter):
            service_filter = ServiceFilter(*service_filter)
        if service_filter.name == "*":
            service_filter.name = entry_name

        if entry_name not in self.entries:
            dependency = compose_instance(DependencyImpl, dependency_args(
                None, service_filter, lifecycle_manager_url, storage_url))
            self.entries[entry_name] = dependency
            self.ec_producer.update("entries", len(self.entries))

    def _ec_producer_change_handler(self, command, entry_name, entry_value):
        if entry_name == "log_level":
            self.logger.setLevel(str(entry_value).upper())

    def exit(self):
        aiko.process.terminate()

    def _get_entry_records(self, entry_name=None, level=0):
        entry_records = []
        for entry_key, entry in self.entries.items():
            if entry_name is None or entry_name == entry_key:
            # TODO: "(dn (None * * * * (a=b c=d)) 0: 0:)"         # None --> 0:
                entry = str(entry).replace("None", "0:")
                entry_record = f"{level} {entry_key} {entry}"
                entry_records.append(entry_record)
        entry_records.sort()
        return entry_records

    def is_type(self, type_name):
        if type_name.lower() == "category":
            return True
        return self.dependency.is_type(self, type_name)

# Note: "recursive" only needed for HyperSpace, not Category, ignore it here

    def list(self, topic_path_response, entry_name=None, recursive=False):
        entry_records = self._get_entry_records(entry_name)
        self._list_publish(topic_path_response, entry_records)

    def _list_publish(self, topic_path_response, entry_records):
        aiko.message.publish(
            topic_path_response, f"(item_count {len(entry_records)})")
        for entry_record in entry_records:
            aiko.message.publish(
                topic_path_response, f"(response {entry_record})")

    def remove(self, entry_name):
        if entry_name in self.entries:
            del self.entries[entry_name]
            self.ec_producer.update("entries", len(self.entries))

    def __repr__(self):
        return self.dependency.__repr__(self)

    @classmethod
    def list_command(cls,
        actor_name, entry_name, long_format, recursive, protocol):

        #TODO: Turn this into @dataclass(es)
        # Record types: A) new child Category, B) current Catefory Entry
        # A) record[0]: -indent [1]: category_name
        # B) record[0]:  indent [1]: entry_name, [2] ServiceFilter LCM Storage
        # class ServiceFilter:    see src/aiko_services/main/service.py
        # class LifeCycleManager: see src/aiko_services/main/lifecycle.py
        # class StorageFile:      see src/aiko_services/main/storage.py

        def response_handler(response):
            if len(response):
                if long_format:
                    output =  "Name: (Service) LifeCycleManager, Storage\n"
                    output += "      (Topic Name Protocol Transport Owner Tags)"
                else:
                    output = "Name: Protocol Owner"

                for record in response:
                    indent = int(record[0]) * 2     # Category depth formatting
                    if indent < 0:                  # new child Category
                        indent = -indent
                        output += f"\n\n{' '*indent} {record[1]}/"
                    else:                           # current Category Entry
                        entry_name = record[1]
                        if record[2][0]:
                            service_filter = ServiceFilter(*record[2][0])
                            protocol = service_filter.protocol
                            after_slash = protocol.rfind("/") + 1
                            service_filter.protocol = protocol[after_slash:]
                        else:
                            service_filter = ServiceFilter()
                        if long_format:
                            lifecycle_manager_url = record[2][1]
                            storage_url = record[2][2]
                            output += f"\n   {' '*indent}{entry_name}: "  \
                                      f"{service_filter} "  \
                                      f"{lifecycle_manager_url}, "  \
                                      f"{storage_url}"
                        else:
                            name = service_filter.name
                            if name is None or name == "*" or name == entry_name:
                                name = ""
                                output += f"\n   {' '*indent}{entry_name}: "  \
                                          f"{service_filter.protocol} "  \
                                          f"{service_filter.owner}  {name}"
            else:
                output = "No category entries"
            print(output)

        do_request(Category,
            ServiceFilter(name=actor_name, protocol=protocol),
            lambda actor: actor.list(_RESPONSE_TOPIC, entry_name, recursive),
            response_handler, _RESPONSE_TOPIC, terminate=True)

    def update(self, entry_name, service=None,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):

    # Prevent ServiceFilter() from overriding "name" with "hostname()"
        if service_filter and not isinstance(service_filter, ServiceFilter):
            service_filter_name_null = service_filter[1] is None
            service_filter = ServiceFilter(*service_filter)
            if service_filter_name_null:
                service_filter.name = None

        if entry_name in self.entries:
            entry = self.entries[entry_name]
            if service_filter:
                if service_filter.name:
                    entry.service_filter.name = service_filter.name
                if service_filter.protocol:
                    entry.service_filter.protocol = service_filter.protocol
                if service_filter.transport:
                    entry.service_filter.transport = service_filter.transport
                if service_filter.owner:
                    entry.service_filter.owner = service_filter.owner
                if service_filter.tags:
                    entry.service_filter.tags = service_filter.tags

            if lifecycle_manager_url:
                entry.lifecycle_manager_url = lifecycle_manager_url

            if storage_url:
                entry.storage_url = storage_url

# --------------------------------------------------------------------------- #

@click.group()

def main():
    """Create, Read/List, Update and Destroy Categories"""
    pass

@main.command(name="add", no_args_is_help=True)
@click.argument("category_name", type=str, required=True, default=None)
@click.argument("entry_name", type=str, required=True, default=None)
@click.option("--service_name", "-n", type=str, default="*",
    help="Service name")
@click.option("--protocol", "-p", type=str, default="*",
    help="Service protocol")
@click.option("--transport", "-tr", type=str, default="*",
    help="Service transport")
@click.option("--owner", "-o", type=str, default="*",
    help="Service owner")
@click.option("--tags", "-t", type=str, multiple=True, default=None,
    help="Service tags")
@click.option("--lifecycle_manager_url", "-lcm", type=str, default=None,
    help="LifeCycleManager URL")
@click.option("--storage_url", "-s", type=str, default=None,
    help="Storage URL")

def add_command(category_name, entry_name,
    service_name, protocol, transport, owner, tags,
    lifecycle_manager_url, storage_url):

    """Add Category Entry

    aiko_category add CATEGORY_NAME ENTRY_NAME

    \b
    • CATEGORY_NAME: Category name
    • ENTRY_NAME: Entry name
    """

    tags = tags if tags else []                 # Assign default tags value
    service_filter = ServiceFilter(             # TODO: Or use ServiceFields ??
        "*", service_name, protocol, transport, owner, tags)

    do_command(Category,
        ServiceFilter(name=category_name, protocol=CATEGORY_PROTOCOL),
        lambda category: category.add(entry_name,
            service_filter, lifecycle_manager_url, storage_url), terminate=True)
    aiko.process.run()

@main.command(name="exit", no_args_is_help=True)
@click.argument("category_name", required=True)

def exit_command(category_name):
    """Exit Category

    aiko_category exit CATEGORY_NAME

    \b
    • CATEGORY_NAME: Category name
    """

    do_command(Category,
        ServiceFilter(name=category_name, protocol=CATEGORY_PROTOCOL),
        lambda category: category.exit(), terminate=True)
    aiko.process.run()

@main.command(name="list", no_args_is_help=True)
@click.argument("category_name", type=str, required=True, default=None)
@click.argument("entry_name", type=str, required=False, default=None)
@click.option("--long_format", "-l", is_flag=True,
    help="Long format with Service, LifeCycleManager URL, Storage URL")

def list_command(category_name, entry_name, long_format):
    """List Category Entries

    aiko_category list CATEGORY_NAME [ENTRY_NAME]

    \b
    • CATEGORY_NAME: Category name
    • ENTRY_NAME: Entry name
    """

    CategoryImpl.list_command(
        category_name, entry_name, long_format, CATEGORY_PROTOCOL)
    aiko.process.run()

@main.command(name="remove", no_args_is_help=True)
@click.argument("category_name", type=str, required=True, default=None)
@click.argument("entry_name", type=str, required=True, default=None)

def remove_command(category_name, entry_name):
    """Remove Category Entry

    aiko_category remove CATEGORY_NAME ENTRY_NAME

    \b
    • CATEGORY_NAME: Category name
    • ENTRY_NAME: Entry name
    """

    do_command(Category,
        ServiceFilter(name=category_name, protocol=CATEGORY_PROTOCOL),
        lambda category: category.remove(entry_name), terminate=True)
    aiko.process.run()

@main.command(name="run", no_args_is_help=True)
@click.argument("category_name", required=True)

def run_command(category_name):
    """Run Category

    aiko_category run CATEGORY_NAME

    \b
    • CATEGORY_NAME: Category name
    """

    tags = ["ec=true"]       # TODO: Add ECProducer tag before add to Registrar
    init_args = actor_args(category_name, None, None, CATEGORY_PROTOCOL, tags)
    category = compose_instance(CategoryImpl, init_args)
    aiko.process.run()

@main.command(name="update", no_args_is_help=True)
@click.argument("category_name", type=str, required=True, default=None)
@click.argument("entry_name", type=str, required=True, default=None)
@click.option("--service_name", "-n", type=str, default="0:",     # None --> 0:
    help="Service name")
@click.option("--protocol", "-p", type=str, default="0:",         # None --> 0:
    help="Service protocol")
@click.option("--transport", "-tr", type=str, default="0:",       # None --> 0:
    help="Service transport")
@click.option("--owner", "-o", type=str, default="0:",            # None --> 0:
    help="Service owner")
@click.option("--tags", "-t", type=str, multiple=True, default=None,
    help="Service tags")
@click.option("--lifecycle_manager_url", "-lcm", type=str, default=None,
    help="LifeCycleManager URL")
@click.option("--storage_url", "-s", type=str, default=None,
    help="Storage URL")

def update_command(category_name, entry_name,
    service_name, protocol, transport, owner, tags,
    lifecycle_manager_url, storage_url):

    """Update Category Entry

    aiko_category update CATEGORY_NAME ENTRY_NAME

    \b
    • CATEGORY_NAME: Category name
    • ENTRY_NAME: Entry name
    """

    tags = tags if tags else []                 # Assign default tags value
    service_filter = ServiceFilter(             # TODO: Or use ServiceFields ??
        "*", service_name, protocol, transport, owner, tags)

    do_command(Category,
        ServiceFilter(name=category_name, protocol=CATEGORY_PROTOCOL),
        lambda category: category.update(entry_name, None,
            service_filter, lifecycle_manager_url, storage_url), terminate=True)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

import numpy as np
import shapely.geometry as geom
from PIL.Image import Image
from pydantic import BaseModel, ConfigDict, Field, field_validator

from highlighter.core import GQLBaseModel
from highlighter.core.utilities import stringify_if_not_null

from ...core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    polygon_from_mask,
)
from .base_models import EAVT
from .datum_source import DatumSource
from .observation import Observation
from .validated_set import ValidatedSet


class AnnotationCrop(BaseModel):
    content: Any
    annotation_id: UUID
    entity_id: Optional[UUID] = None


_sentinel = object()


class Annotation(GQLBaseModel):
    id: UUID = Field(..., default_factory=uuid4)
    _entity: Optional[Any] = None  # Optional[Entity], not imported to avoid circular imports

    location: Optional[Union[geom.Polygon, geom.MultiPolygon, geom.LineString, geom.Point]] = None
    data_file_id: Optional[UUID] = None
    observations: ValidatedSet
    track_id: Optional[UUID] = None
    correlation_id: Optional[UUID] = None

    # Frame ID in datum_source
    datum_source: DatumSource = Field(..., default_factory=lambda: DatumSource(confidence=1))
    occurred_at: Optional[datetime] = None  # FIXME make non-optional

    model_config = ConfigDict(
        arbitrary_types_allowed=True, extra="ignore"
    )  # Required for shapely geometry types

    def __init__(self, **kwargs):
        observations = ValidatedSet()
        if "observations" in kwargs and kwargs["observations"] is not None:
            for obs in kwargs["observations"]:
                observations.append(obs)
        kwargs["observations"] = observations
        super().__init__(**kwargs)
        if "entity" in kwargs:
            self.entity = kwargs["entity"]  # Trigger setter
        # Validate observations passed to constructor
        self.observations.validator = self.validate_add_observation
        for observation in self.observations:
            self.validate_add_observation(observation)

    @property
    def entity(self) -> Optional["Entity"]:
        return self._entity

    @entity.setter
    def entity(self, new_entity: Optional["Entity"]):
        from .entity import Entity  # Import here to avoid circular imports

        if not isinstance(new_entity, Optional[Entity]):
            raise TypeError(f"Expected Entity, got '{type(new_entity).__qualname__}'")

        # Check if update is allowed
        if new_entity is not None and self.entity is not None and new_entity != self.entity:
            raise ValueError(
                f"Annotation is still associated with entity {self.entity.id}. "
                "Dissiaociate via `self.entity = None` before "
                "associating with a different entity."
            )

        # Do the update
        if self.entity is not None:
            self.entity.annotations.remove(self)
        if new_entity is not None and self not in new_entity.annotations:
            new_entity.annotations.add(self)
        self._entity = new_entity

    def validate_add_observation(self, new_observation: Observation):
        if not isinstance(new_observation, Observation):
            raise TypeError(f"Expected Observation, got {type(new_observation).__qualname__}")

        # Check if update is allowed
        if new_observation._global_observation_entity is not None:
            raise ValueError(
                "Cannot add observation that references a different entity. "
                f"Dissasociate the observation from entity {new_observation.entity.id} "
                f"via `observation.entity = None` before adding it to annotation {self.id}"
            )
        if new_observation._annotation is not None and new_observation._annotation != self:
            raise ValueError(
                "Cannot add observation that references a different annotation. "
                f"Dissasociate the observation from annotation {new_observation._annotation.id} "
                f"via `observation.annotation = None` before adding it to annotation {self.id}"
            )

        # Do associated updates
        new_observation._annotation = self

    def get_or_create_entity(self) -> "Entity":
        from .entity import Entity

        if self.entity is not None:
            return self.entity
        entity = Entity()
        entity.annotations.add(self)
        return entity

    @classmethod
    def from_points(
        cls,
        points: List[Tuple[int, int]],
        confidence: float,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
        occurred_at: Optional[datetime] = None,
    ):
        datum_source = DatumSource(
            confidence=confidence,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
        )

        anno = cls(
            location=geom.Polygon(points),
            observations=observations,
            track_id=track_id,
            correlation_id=correlation_id,
            data_file_id=data_file_id,
            datum_source=datum_source,
            occurred_at=occurred_at,
        )
        return anno

    @classmethod
    def from_left_top_right_bottom_box(
        cls,
        box: Tuple[int, int, int, int],
        confidence: float,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
        occurred_at: Optional[datetime] = None,
    ):
        x0, y0, x1, y1 = box
        return cls.from_points(
            [(x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)],
            confidence,
            data_file_id=data_file_id,
            observations=observations,
            track_id=track_id,
            correlation_id=correlation_id,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
            occurred_at=occurred_at,
        )

    @classmethod
    def from_mask(
        cls,
        mask: np.array,
        confidence: float,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
        occurred_at: Optional[datetime] = None,
        **polygon_from_mask_kwargs,
    ):

        points = polygon_from_mask(mask, **polygon_from_mask_kwargs)
        points = [(x, y) for x, y in zip(points[:-1:2], points[1::2])]
        # Shapely expects a closed
        points.append(points[0])
        # breakpoint()

        return cls.from_points(
            points,
            confidence,
            data_file_id=data_file_id,
            observations=observations,
            track_id=track_id,
            correlation_id=correlation_id,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
            occurred_at=occurred_at,
        )

    def observations_where(
        self, *, attribute_id: UUID | None = None, value: Any = _sentinel
    ) -> List[Observation]:
        obs_where = []
        for o in self.observations:
            if attribute_id is not None and o.attribute_id == attribute_id:
                if value is _sentinel or o.value == value:
                    obs_where.append(o)
        return obs_where

    def crop(self, crop_args: "CropArgs", data_sample: "DataSample") -> AnnotationCrop:
        from ...datasets.cropping import crop_rect_from_poly

        if self.data_file_id is None:
            raise ValueError("Cannot crop an Annotation when data_file_id is None")
        if self.location is None:
            raise ValueError("Cannot crop an Annotation when location is None")
        if not isinstance(self.location, geom.Polygon):
            raise ValueError(f"Cannot crop an Annotation when location is {type(self.location)}")

        if "image" not in data_sample.content_type:
            raise ValueError(
                f"Cannot crop data_file with content_type '{data_sample.content_type}', must be 'image'"
            )

        if not isinstance(data_sample.content, (np.ndarray, Image)):
            raise ValueError(
                f"Cannot crop data_file with content '{type(data_sample.content).__qualname__}', must be (PIL.Image|np.array)"
            )

        cropped_image = crop_rect_from_poly(data_sample.content, self.location, crop_args)
        return AnnotationCrop(content=cropped_image, entity_id=self.entity.id, annotation_id=self.id)

    def to_deprecated_pixel_location_eavt(self) -> EAVT:
        if self.entity is None:
            raise ValueError("Entity must exist to convert Annotation to deprecated EAVT")
        return EAVT.make_pixel_location_eavt(
            entity_id=self.entity.id,
            location_points=self.location,
            confidence=self.datum_source.confidence,
            time=datetime.now(timezone.utc),
            frame_id=self.datum_source.frame_id,
        )

    @field_validator("location")
    @classmethod
    def validate_geometry(cls, v):
        if v is not None:
            assert v.is_valid, f"Invalid Geometry: {v}"
        return v

    def serialize(self):
        return {
            "id": str(self.id),
            "entity_id": str(self.entity.id) if self.entity else None,
            "location": self.location.wkt if self.location is not None else None,
            "observations": [o.serialize() for o in self.observations],
            "track_id": str(self.track_id),
            "correlation_id": str(self.correlation_id),
            "data_file_id": str(self.data_file_id),
            "datum_source": self.datum_source.serialize(),
        }

    def to_json(self):
        data = self.model_dump()
        data["datum_source"] = self.datum_source.model_dump()
        data["id"] = stringify_if_not_null(data["id"])
        data["entity_id"] = stringify_if_not_null(data.get("entity", {}).get("id"))
        data["location"] = data["location"].wkt if data["location"] is not None else None
        data["observations"] = [d.to_json() for d in data["observations"]]

        data["track_id"] = stringify_if_not_null(data["track_id"])
        data["data_file_id"] = stringify_if_not_null(data["data_file_id"])
        data["correlation_id"] = stringify_if_not_null(data["correlation_id"])
        data["occurred_at"] = self.occurred_at.isoformat() if self.occurred_at is not None else None
        return data

    def gql_dict(self) -> Dict:
        try:
            object_class_observation = [
                observation
                for observation in self.observations
                if observation.attribute_id == OBJECT_CLASS_ATTRIBUTE_UUID
            ][-1]
        except IndexError:
            raise ValueError(
                "Annotation must have an object-class observation in order to submit to Highlighter"
            )

        if isinstance(self.location, geom.Polygon):
            data_type = "polygon"
        elif isinstance(self.location, geom.LineString):
            data_type = "line"
        elif isinstance(self.location, geom.Point):
            data_type = "point"
        else:
            data_type = "polygon"
        result = {
            "objectClassUuid": stringify_if_not_null(object_class_observation.value),
            "location": self.location.wkt if self.location is not None else None,
            "confidence": self.datum_source.confidence,
            "dataType": data_type,
            "correlationId": stringify_if_not_null(self.correlation_id),
            "frameId": self.datum_source.frame_id,
            "trackId": stringify_if_not_null(self.track_id),
            "entityId": str(self.entity.id) if self.entity else None,
            "dataFileId": stringify_if_not_null(self.data_file_id),
            "uuid": stringify_if_not_null(self.id),
        }
        return result

    def to_deprecated_observations(self) -> List[Observation]:
        """
        Convert to the deprecated "set of observations" representation
        where pixel locations are represented as observations rather than annotations
        """
        observations = list(self.observations)
        if self.location is not None:
            observation = Observation.make_pixel_location_observation(
                value=self.location,
                confidence=self.datum_source.confidence,
                occurred_at=self.occurred_at,
                frame_id=self.datum_source.frame_id,
            )
            observation.annotation = self
            observations.append(observation)
        return observations



import json
from typing import Dict, Optional, Tuple, Union

import litellm
from litellm import ModelResponse, completion
from litellm.caching.caching import Cache

from highlighter.agent.capabilities import Capability, StreamEvent

__all__ = ["LLM"]


class LLM(Capability):
    """
    Parameters used by this capability:


    model: str: name of the model to use. Default: "claude-3-haiku-20240307"

    prompt_template: str: The prompt to use the context of the process_frame 'text'
        param will be inserted at {{PLACEHOLDER}}

    prompt_template_path: str: If used the `prompt_template` will be overwirtten
        with the contents of the template file.

    system_prompt_template: str: The system prompt template to use

    system_prompt_template_path: str: If used the `system_prompt_template` will be overwirtten
        with the contents of the template file.

    mock_response: bool: mock responses from llm, Defaut `true`

    num_retries: int: Number of retries. Default: 8

    completion_kwargs": {"temperature": 0.5}
        timeout: float | int | None = None
        temperature: float | None = None
        top_p: float | None = None
        n: int | None = None
        stream: bool | None = None
        stop: Unknown | None = None
        max_tokens: int | None = None
        presence_penalty: float | None = None
        frequency_penalty: float | None = None
        logit_bias: dict[Unknown, Unknown] | None = None
        user: str | None = None
        response_format: dict[Unknown, Unknown] | None = None
        seed: int | None = None
        tools: List[Unknown] | None = None
        tool_choice: str | None = None
        logprobs: bool | None = None
        top_logprobs: int | None = None
        deployment_id: Unknown | None = None
        extra_headers: dict[Unknown, Unknown] | None = None
        functions: List[Unknown] | None = None
        function_call: str | None = None
        base_url: str | None = None
        api_version: str | None = None
        api_key: str | None = None
        model_list: list[Unknown] | None = None
    """

    class InitParameters(Capability.InitParameters):
        cache_system_prompt: bool = False
        completion_kwargs: dict = {}
        mock_response: bool = True
        model: str = "claude-3-haiku-20240307"
        num_retries: int = 8
        prompt_template: Optional[str] = None
        system_prompt_template: Optional[str] = None

    class StreamParameters(InitParameters):
        prompt_template_path: Optional[str] = None
        system_prompt_template_path: Optional[str] = None

    def __init__(self, context):
        super().__init__(context)
        self.prompt_template: str = self.init_parameters.prompt_template
        self.system_prompt_template: Optional[str] = self.init_parameters.system_prompt_template

        litellm.cache = Cache(type="disk")
        litellm.success_callback = [self.log_success_event]
        litellm.failure_callback = [self.log_failure_event]
        litellm._logging._disable_debugging()

    def start_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        parameters = self.stream_parameters(stream.stream_id)

        if parameters.prompt_template_path is not None:
            with open(parameters.prompt_template_path, "r") as f:
                self.prompt_template = f.read()
        if not self.prompt_template:
            raise ValueError("Must supply 'prompt_template'")

        if parameters.system_prompt_template_path is not None:
            with open(parameters.system_prompt_template_path, "r") as f:
                self.system_prompt_template = f.read()
        if not self.system_prompt_template:
            raise ValueError("Must supply 'system_prompt_template'")

        if parameters.strategy is None:
            raise ValueError("Agent parameter 'strategy' is not set.")

        elif parameters.strategy in ["select_all", "pass", "completion"]:
            pass

        else:
            raise ValueError("Agent parameter 'strategy' needs valid value.")

        return StreamEvent.OKAY, None

    def process_frame(self, stream, text: str) -> Tuple[StreamEvent, Union[Dict, str]]:
        """

        text: str

        """
        parameters = self.stream_parameters(stream.stream_id)
        if parameters.strategy == "select_all":
            return StreamEvent.OKAY, {"text": "select_all", "input_text": text}
        elif parameters.strategy == "pass":
            return StreamEvent.OKAY, {"text": "pass", "input_text": text}

        if not isinstance(text, str):
            raise TypeError(f"Invalid content type, expected str got type '{type(text)}' with value '{text}'")

        prompt = self.prompt_template.replace("{{PLACEHOLDER}}", text)

        content = ""

        try:
            messages = []
            system = []

            if self.system_prompt_template:
                if parameters.cache_system_prompt:
                    system.append(
                        {
                            "type": "text",
                            "text": self.system_prompt_template,
                            "cache_control": {"type": "ephemeral"},
                        }
                    )
                else:
                    system.append({"type": "text", "text": self.system_prompt_template})

                messages.append({"content": system, "role": "system"})

            messages.append({"content": prompt, "role": "user"})

            if parameters.mock_response:
                mock_content = json.dumps([{"mock_response": "mock_response_value"}] * 3)
                response = ModelResponse(**{"choices": [{"message": {"content": mock_content}}]})
            else:
                response: ModelResponse = completion(
                    model=parameters.model,
                    messages=messages,
                    num_retries=parameters.num_retries,
                    caching=True,
                    **parameters.completion_kwargs,
                )

            content = response["choices"][0]["message"]["content"]
        except Exception as e:
            message = f"Error while calling Litellm.completion, got: {e}"
            self.logger.error(message)
            return StreamEvent.ERROR, {"diagnostic": message}

        return StreamEvent.OKAY, {"text": content, "input_text": text}

    def write_log(self, response_type, kwargs, response_obj, start_time, end_time):
        time_log_msgs = []
        if start_time is not None and end_time is not None:
            time_log_msgs.append(f"time: {timedelta_to_readable(end_time - start_time)}")

        base_log_messages = []

        if "response_cost" in kwargs:
            base_log_messages.append(f'response_cost: {round(kwargs["response_cost"], 6)}')

        if "cache_hit" in kwargs:
            cache_hit_y_n = "y" if kwargs["cache_hit"] else "n"
            base_log_messages.append(f"response_cache: {cache_hit_y_n}")

        identity_log_messages = []
        response_log_messages = []

        content = []

        if response_obj is not None:
            try:
                content = json.loads(response_obj["choices"][0]["message"]["content"])
                content = content if isinstance(content, list) else [content]
            except:
                raise ValueError(f"Couldn't parse response content: '{response_obj}'")

            if "usage" in response_obj:
                usage = response_obj["usage"]

                if "cache_creation_input_tokens" in usage:
                    prompt_cache_input_tokens = usage["cache_creation_input_tokens"]
                    if prompt_cache_input_tokens > 0:
                        response_log_messages.append(
                            f"prompt_cache_input_tokens: {prompt_cache_input_tokens}"
                        )

                if "cache_read_input_tokens" in usage:
                    cache_read_input_tokens = usage["cache_read_input_tokens"]
                    if cache_read_input_tokens > 0:
                        response_log_messages.append(
                            f"prompt_cache_read_input_tokens: {cache_read_input_tokens}"
                        )

        identity_log_messages = [json.dumps(content[0])]

        self.logger.info(
            f"LLM {response_type} "
            + " ".join(
                map(str, identity_log_messages + time_log_msgs + base_log_messages + response_log_messages)
            )
        )

    def log_success_event(self, kwargs, response_obj, start_time, end_time):
        self.write_log("success", kwargs, response_obj, start_time, end_time)

    def log_failure_event(self, kwargs, response_obj, start_time, end_time):
        self.write_log("failure", kwargs, response_obj, start_time, end_time)


def timedelta_to_readable(time_delta):
    days = time_delta.days
    hours, remainder = divmod(time_delta.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    milliseconds = time_delta.microseconds // 1000  # Convert microseconds to milliseconds

    time_parts = []
    if days > 0:
        time_parts.append(str(days))
    if hours > 0 or days > 0:
        time_parts.append(f"{str(hours)}hr")
    if minutes > 0 or hours > 0 or days > 0:
        time_parts.append(f"{str(minutes)}min")
    if seconds > 0 or minutes > 0 or hours > 0 or days > 0:
        time_parts.append(f"{str(seconds)}s")
    if milliseconds > 0 or len(time_parts) > 0:
        time_parts.append(f"{str(milliseconds)}ms")

    return " ".join(time_parts)



import logging
from typing import Any, Dict, List, Optional

from ..core import GQLBaseModel, LabeledUUID, paginate
from .base_models import ObjectClass, ObjectClassTypeConnection

__all__ = [
    "read_object_classes",
    "create_object_classes",
]

logger = logging.getLogger(__name__)


def read_object_classes(
    client,
    workflow_id: Optional[int] = None,
    **kwargs,
):
    """See graphiql docs for objectClassConnection for current valid arguments"""
    if workflow_id is None:
        return paginate(client.objectClassConnection, ObjectClassTypeConnection, **kwargs)
    else:

        class ProjectType(GQLBaseModel):
            object_classes: List[ObjectClass]

        return client.project(return_type=ProjectType, id=workflow_id).object_classes


def create_object_classes(
    client,
    names: List[str],
) -> Dict[str, LabeledUUID]:
    """Create object classes in a Highlighter account

    Returns a mapping from the names to the Highlighter ObjectClass UUID

    Note:
      Will check for existing names (case incentive) and use the existing
      ObjectClass UUID as necessary
    """

    existing_object_classes = {o.name.lower(): o.uuid for o in read_object_classes(client)}

    class ReturnType(GQLBaseModel):
        errors: Any = None
        object_class: ObjectClass

    name_to_id_lookup = {}
    for name in names:
        _name = name.lower()
        if name in existing_object_classes:
            logger.debug(f"Found existing object class: {_name}")
            name_to_id_lookup[_name] = LabeledUUID(existing_object_classes[_name], label=_name)
        else:
            result = client.createObjectClass(
                return_type=ReturnType,
                name=_name,
                default=False,
            ).object_class
            name_to_id_lookup[result.name] = LabeledUUID(result.uuid, label=result.name)

    return name_to_id_lookup



# Declaration order is based on the dependency on static references
#
# To Do
# ~~~~~
# - None, yet !

from .elements import (
    all_outputs, evaluate, evaluate_condition, evaluate_define, Expression
)



# To Do
# ~~~~~

import sys

import aiko_services as aiko

__all__ = ["HLDataSchemePipe"]

# --------------------------------------------------------------------------- #


class HLDataSchemePipe(aiko.DataScheme):

    def create_sources(self, stream, data_sources, frame_generator=None, use_create_frame=True):

        source_buffers = []
        for _ in data_sources:
            try:
                source_buffers.append(sys.stdin.buffer)
            except Exception as e:
                diagnostic = f'Error loading file, "{e}"'
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        pipeline_element = self.pipeline_element
        if use_create_frame:
            """
            - Not sure how to determine the frame_data keys ahead of time.
            - Not sure how this makes sense for DataSource Elements that
              produce many frames for one file, ie VideoFileRead
            """
            NotImplementedError()
        else:
            stream.variables["source_buffers"] = source_buffers
            rate, _ = pipeline_element.get_parameter("rate", default=None)
            rate = float(rate) if rate else None
            pipeline_element.create_frames(stream, frame_generator, rate=rate)
        return aiko.StreamEvent.OKAY, {}

    def create_targets(self, stream, data_targets):
        return aiko.StreamEvent.OKAY, {}

    def destroy_sources(self, stream):
        self.terminate = True

    def frame_generator(self, stream, frame_id):
        raise ValueError("Use the DataSourceCapability frame_generator")


aiko.DataScheme.add_data_scheme("hlpipe", HLDataSchemePipe)

# --------------------------------------------------------------------------- #



from .annotation import *
from .base_models import *
from .data_file import *
from .datum_source import *
from .entity import *
from .image import *
from .object_class import *
from .observation import *
from .page_info import *
from .submission import *
from .user import *



# Usage
# ~~~~~
# - aiko_pipeline create pipelines/pipeline_expression.json -fd "()" -ll debug
#
# To Do
# ~~~~~
# - Expression evaluation support ...
#   - Dictionary or List access, e.g variable[1] or variable[index]
#     - Requires parser improvement for list[index], e.g variable[index]
#   - Extensible list of functions, e.g random(), stack push() / pop()
#     - Requires parser improvement for function calls, e.g random()
#
# - Common system calls, e.g /proc, filesystem, network, etc
# - Update Pipeline Stream Event --> State, implement StateMachine
# - Alter Pipeline Graph control flow
#   - Conditionals

import re
from typing import Tuple

import aiko_services as aiko
from aiko_services.main.utilities import *

__all__ = [
    "all_outputs", "evaluate", "evaluate_conition", "evaluate_define",
    "Expression"
]

# --------------------------------------------------------------------------- #

def all_outputs(pipeline_element, stream):
    frame = stream.frames[stream.frame_id]
    outputs = {}
    for output_definition in pipeline_element.definition.output:
        output_name = output_definition["name"]
        outputs[output_name] = frame.swag[output_name]
    return outputs

# --------------------------------------------------------------------------- #
# Expression PipelineElement enables modifying process_frame()
# input and output arguments, via "define", "delete" and "rename"
# commands, which are specified as Expression parameters ...
#
# - define: "(argument_0 expression_0) (argument_1 expression_1) ..."
# - delete: "(argument_0 argument_1 ...)"
# - rename: "(from_0 to_0) (from_1 to_1)"
#
# - expression: expression_0 operator expression_1 ...
# - expression: integer, float, string, list, dictionary, argument_name
# - operator:   + - * / < <= > >=

def evaluate(expression, arguments={}):
    expression = expression.strip()

    if expression in arguments:  # Is an argument ?
        return arguments[expression]

    if re.fullmatch(r"\d+(\.\d+)?", expression):  # Is a number ?
        return float(expression) if "." in expression else int(expression)

    if re.fullmatch(r"\".*\"|\'.*\'" , expression) or  \
        re.fullmatch(r"[a-zA-Z_][a-zA-Z0-9_]*", expression):
        return expression.strip("\"'")  # If present, remove quotes

    if re.fullmatch(r"\[.*\]", expression):
        return eval(expression)  # Convert "list" to a [list]

    match = re.search(r"(\+|\-|\*|/|\<|\<=|/>|\>=)", expression)  # operators
    if match:
        operator = match.group()
        left_expression = expression[:match.start()].strip()
        right_expression = expression[match.end():].strip()

        left_value = evaluate(left_expression, arguments)
        right_value = evaluate(right_expression, arguments)

        left_value = parse_number(left_value, left_value)     # coerce to number
        right_value = parse_number(right_value, right_value)  # ditto

        if isinstance(left_value, (int, float)) and  \
             isinstance(right_value, (int, float)):
            if operator == "+":
                return left_value + right_value
            elif operator == "-":
                return left_value - right_value
            elif operator == "*":
                return left_value * right_value
            elif operator == "/":
                if right_value == 0:
                    raise ValueError("Invalid division by zero")
                return left_value / right_value
            elif operator == "<":
                return left_value < right_value
            elif operator == "<=":
                return left_value <= right_value
            elif operator == ">":
                return left_value > right_value
            elif operator == ">=":
                return left_value >= right_value

        elif operator == "+" and  \
             isinstance(left_value, list) and isinstance(right_value, list):
            return left_value + right_value

        elif operator == "+" and  \
             isinstance(left_value, (int, float, str)) and  \
             isinstance(right_value, (int, float, str)):
            return f"{left_value}{right_value}"

    return expression  # Treat everything else as a "string"
#   raise ValueError(f"Invalid expression: {expression}")

def evaluate_condition(expressions, swag, logger=None):
    results = []
    if isinstance(expressions, list):
        for expression in expressions:
            if logger:
                logger.debug(f"Expression: {expression}")
            if isinstance(expression[0], str):
                expression[0] = evaluate(expression[0], swag)
            results.append(True if expression[0] else False)
    return all(results)

def evaluate_define(expressions, swag, logger=None, name="Define"):
    if isinstance(expressions, list):
        for expression in expressions:
            if logger:
                logger.debug(f"{name}: {expression}")
            if isinstance(expression[1], str):
                expression[1] = evaluate(expression[1], swag)
            swag[expression[0]] = expression[1]

class Expression(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("expression:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        swag = stream.frames[stream.frame_id].swag

        for command_name in ["define", "delete", "rename"]:
            command, found = self.get_parameter(command_name)
            if found:
                expressions = parse(command, car_cdr=False)
                if command_name == "define":
                    evaluate_define(expressions, swag, self.logger)
                if command_name == "delete" and isinstance(expressions, list):
                    for name in expressions:
                        self.logger.debug(f"Delete: {name}")
                        if name in swag:
                            del swag[name]
                if command_name == "rename" and isinstance(expressions, list):
                    for name in expressions:
                        self.logger.debug(f"Rename: {name}")
                        if len(name) == 2 and name[0] in swag:
                            swag[name[1]] = swag.pop(name[0])
        return aiko.StreamEvent.OKAY, all_outputs(self, stream)

# --------------------------------------------------------------------------- #
